/*
 *  3D Studio to OpenGL Converter 
 *  by David Farrell
 *    &
 *  3D Studio to OpenGL File Converter
 *  by David Farrell, fdavid@cyberramp.net
 */

/*
 *  3D Studio object reader
 *    &
 *  OpenGL Renderer
 */
 
/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */


#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <new.h>
#include <string.h>
#include <math.h>
#include "mygl.h"
#include "3dsLoad.h"
#include "Textures.h"
#include "Scene.h"
#include "Log.h"
#define WANT_NEW
#include "porting.h"

#include "setup.h"



#define DISABLE_THRESHOLD


#define MAXNAMESIZE 50
#define VERTEXBLOCKSIZE 1024


#define JOB_DONE           0
#define DEGENERATE_SURFACE 1

enum {
	CHUNK_RGB1      	= 0x0010,  /* 3 floats of RGB */
	CHUNK_RGB2      	= 0x0011,  /* 3 bytes of RGB */
	CHUNK_RGB3        = 0x0012,  /* What is this? */
	CHUNK_AMOUNT      = 0x0030,
	CHUNK_MAIN      	= 0x4D4D,
	CHUNK_OBJMESH     = 0x3D3D,
	CHUNK_BKGCOLOR    = 0x1200,
	CHUNK_AMBCOLOR  	= 0x2100,
	CHUNK_OBJBLOCK  	= 0x4000,
	CHUNK_TRIMESH   	= 0x4100,
	CHUNK_VERTLIST    = 0x4110,
	CHUNK_FACELIST    = 0x4120,
	CHUNK_FACEMAT     = 0x4130,
	CHUNK_MAPLIST     = 0x4140,
	CHUNK_SMOOLIST    = 0x4150,
	CHUNK_TRMATRIX    = 0x4160,
	CHUNK_LIGHT     	= 0x4600,
	CHUNK_SPOTLIGHT   = 0x4610,
	CHUNK_CAMERA    	= 0x4700,
	CHUNK_MATERIAL  	= 0xAFFF,
	CHUNK_MATNAME   	= 0xA000,
	CHUNK_AMBIENT   	= 0xA010,
	CHUNK_DIFFUSE   	= 0xA020,
	CHUNK_SPECULAR  	= 0xA030,
	CHUNK_SHININESS       = 0xA040,
	CHUNK_SHINE_STRENGTH  = 0xA041,
	CHUNK_TRANSPARENCY    = 0xA050,
	CHUNK_DOUBLESIDED     = 0xA081,
	CHUNK_TEXTURE   	= 0xA200,
	CHUNK_BUMPMAP   	= 0xA230,
	CHUNK_MAPFILENAME   	= 0xA300,
	CHUNK_MAPOPTIONS      = 0xA351,
	CHUNK_KEYFRAMER 	= 0xB000,
	CHUNK_FRAMES          = 0xB008,
	CHUNK_StartOfObjectInfo = 0xB002,
	CHUNK_NameHierarchyDescriptor = 0xB010
};


typedef unsigned char byte;     /* 8-bit */
typedef short word;             /* 16-bit */
typedef unsigned short uword;   /* 16-bit */
typedef int   dword;            /* 32-bit */
typedef float float32;          /* 32-bit float */

typedef struct {
	word p0,p1,p2;
	word flags;
} H3dsFace;

typedef struct {
	float32 x;
	float32 y;
	float32 z;
} H3dsVert;

typedef struct {
	float32 u;
	float32 v;
} H3dsMap;

/* Material structures */
typedef struct {
	char name[MAXNAMESIZE];
	int  NumFaces;
	uword *faces;                   /* Changed from word to uword */
} H3dsMatList;

typedef struct {
	float32 red;
	float32 green;
	float32 blue;
} H3dsRGBColor;

typedef struct {
	char texturename[MAXNAMESIZE];
	word options;
} H3dsTexture;

typedef struct {
	char name[MAXNAMESIZE];
	H3dsRGBColor ambient;
	H3dsRGBColor diffuse;
	H3dsRGBColor specular;
	word mat_type;
	int doublesided;
	H3dsTexture texture;
	int offset;
	float32 alpha;
} H3dsMat;

/* Object and scene structures */
typedef struct {
	char name[MAXNAMESIZE]; /* object name, zero terminated */
	int NumFaces;           /* number of faces in face list */
	int NumVerts;           /* number of vertices in vertex list */
	int NumMaps;            /* number of mapping coordinates in map list */
	int matrix;             /* 1 if transformation matrix is loaded */
	int NumBindings;        /* number of binding materials */
	H3dsFace *facelist;     /* array of face indices */
	H3dsVert *vertlist;     /* array of vertices */
	H3dsMap  *maplist;      /* array of texture coordinates */
	float32 TraMatrix[3*4]; /* 3*3 rotation matrix, 3*1 translation matrix */
	H3dsMatList *bindings;  /* Binds faces to materials */
	int offset;             /* Our starting index into outgl's vertex list */
} H3dsMeshObj;

typedef struct {
	int meshobjs;              /* Mumber of meshobjects in meshobjlist */
	H3dsMeshObj *meshobjlist;  /* array of meshobjects */
	int NumMaterials;          /* Number of materials in the scene */
	H3dsMat *material;         /* array of materials */
} H3dsScene;





typedef struct {
	GLfloat red;
	GLfloat green;
	GLfloat blue;
	GLfloat alpha;
} RGB;

typedef struct {
	GLfloat u,v;                       /* Texture Coordinates */
	GLfloat nx, ny, nz;                /* Normal Vector */
	GLfloat x, y, z;                   /* XYZ coordinates for vertex */
} VERTEX;

typedef struct {
	GLfloat n0, n1, n2;                /* Triangle's normal */
} NORMAL;

typedef struct {
	GLuint p0, p1, p2;                 /* Indices into OUTGL's vertex array */
} TRIANGLE;

typedef struct {                     /* A surface is a TRIANGLE but with a */
  GLuint p0, p1, p2;                 /* normal for the triangle attached */
  GLfloat n0, n1, n2;
} SURFACE;

struct _TRIANGLENODE {               /* LList used to keep track of which */
	SURFACE *triangle;                 /* triangles touch which vertices. Links */
	struct _TRIANGLENODE *next;        /* the vertices back to the triangles. */
};
typedef struct _TRIANGLENODE TRIANGLENODE;

typedef struct {
  char name[50];                     /* Material name */
  RGB ambient;                       /* Colors */
  RGB diffuse;
	RGB specular;
	int doublesided;
  int NumTriangles;                  /* # of triangles of this material */
  SURFACE *TriangleIndexList;        /* array of 3-tuple ints which are indexes into VertexList[] */
	TRIANGLE *CrunchedTriangleIndexList;
	char texturename[50];
  GLuint textureID;

	int NumStrips;                     /* Number of different triangle strips */
  GLuint *StripStart;                /* Index into StripPoint of each strip's starting point */
	GLuint *StripLength;               /* The number of points in each strip */
	int NumStripPoints;                /* Number of points */
	GLuint *StripPoint;                /* Indexes into the global vertex list */
} POLYGONLIST;

typedef struct
{
	int NumPolyLists;
	int *array;
} POLYGONLISTARRAY;



class Loader3DS {
private:
	FILE *InFile;
	long size;
	H3dsScene *Scene;
	jmp_buf EnvState;
	int doublesided;
	int alphablending;
	int defaultmaterialwarning;

	int NumVertex;
	VERTEX *VertexList;
	int NumPolyLists;
	POLYGONLIST *PolyList;
	//VERTEX center;
	TRIANGLENODE **CommonTriangleList;
	int VertexListSize;
	int flip;
	int PolygonWinding;
	int wireframe;
	int NumPolyListsArray;
	POLYGONLISTARRAY *PolyListArray;

	int UseThreshold;
	GLfloat cos_threshold;

	void* getmem(int size);
	H3dsMat* EnlargeMatObjArray(void);
	H3dsMatList* EnlargeBindingsArray(H3dsMeshObj *meshobj);
	dword dgetpos(void);
	void dsetpos(dword pos);
	void dread(void *dest, int len);
	H3dsMeshObj *EnlargeMeshArray(void);
	void ReadVertList(H3dsMeshObj *meshobj);
	word ReadFaceList(H3dsMeshObj *meshobj);
	void ReadMapList(H3dsMeshObj *meshobj);
	void ReadTraMatrix(H3dsMeshObj *meshobj);
	void ReadFaceMaterial(H3dsMeshObj *meshobj);
	void ReadDefaultMaterial(H3dsMeshObj *meshobj, word nf);
	void ReadTriMeshBlocks(dword p, char *name);
	void ReadObjBlocks(dword p);
	void ReadRGBFloat(dword p, H3dsRGBColor *rgb);
	void ReadRGBByte(dword p, H3dsRGBColor *rgb);
	void ReadMatChunk(dword p, H3dsMat *mat);
	void ReadShadingChunk(dword p, H3dsRGBColor *rgb);
	void ReadTextureNameChunk(dword p, H3dsTexture *texture);
	void ReadTextureChunk(dword p, H3dsTexture *texture);
	void ReadDoublesidedChunk(int *material_doublesided);
	void ReadAmountChunk(word *amount);
	void ReadAlphaChunk(dword p, float32 *alpha);
	void ReadMaterialBlocks(dword p);
	void ReadObjMeshBlocks(dword p);
	void ReadMainBlocks(dword p);
	H3dsScene *HRead3dsScene(void);
	void HFree3dsScene(void);

	int FindTextureIndex(GLuint tex);
	int TextureIndex(void);
	int FindMaterialIndex(char *name);
	int CalculateSurfaceNormal(SURFACE *);
	int OutGLAddMaterial(RGB ambient, RGB diffuse, RGB specular, char *name, char *texname, int doublesided);
	int OutGLAddVertex(VERTEX v);
	int OutGLAddPolygon(TRIANGLE tri, char *MaterialName);
	void OutGLInitialize(void);
	void FreeData(void);
	int ConvertToOutGL(void);

	double distance(VERTEX *v1, VERTEX *v2);
	GLfloat cos_angle(SURFACE *n1, SURFACE *n2);
	GLfloat angle(SURFACE *triangle, GLuint vertex);
	int OutGLCompileDisplayList(void);
	void MakeTriangleIndexList(void);
	int CalculateVertexNormals(void);
	void LinkVerticesToPolygons(void);
	int OutGLFinishModel(void);
public:
	int DisplayList;
	int Load3ds(char *Name);
	Loader3DS()  : Scene(NULL), VertexList(NULL), CommonTriangleList(NULL),
								 NumVertex(0), PolyList(NULL), NumPolyLists(0),
								 NumPolyListsArray(0), PolyListArray(NULL) { UseThreshold = 0; cos_threshold = -0.266f; };
	~Loader3DS()  { HFree3dsScene(); };
};


















int Loader3DS::Load3ds(char *FileName)
{
  convertFilename(FileName);
  logInfo(mod3dsLoad, "Starting loading file \"%s\".", FileName);
	doublesided = 0;
	defaultmaterialwarning = 0;
	alphablending = 0;

	/* Open & prepare input file */
	InFile = fopen(FileName, "rb");
	if(!InFile) {
		logError(mod3dsLoad, "Failed to open file %s.", FileName);
		return 0;
	}

	if(fseek(InFile, 0, SEEK_END)) {
		logError(mod3dsLoad, "Error seeking in file %s.", FileName);
		fclose(InFile);
		return 0;
	}

	if((size=ftell(InFile)) == -1L) {
		logError(mod3dsLoad, "Error sizing file %s.", FileName);
		fclose(InFile);
		return 0;
	}
	rewind(InFile);

	/* Read in the file and place it in *scene */
	if((Scene = HRead3dsScene()) == 0) {
		//logError(mod3dsLoad, "Failed to parse file %s.", FileName);
		fclose(InFile);
		return 0;
	}
	fclose(InFile);

	/* Convert our internal data structure to outgl's structure */
	if (!ConvertToOutGL())
  {
		HFree3dsScene();
    return 0;
  }

	if (defaultmaterialwarning) {
    logWarning(mod3dsLoad, "DEFAULT MATERIAL WARNING\n"
                           "This model uses default materials. 3D Studio saves no\n"
		                       "information about default materials, so View3DS substitutes\n"
		                       "an all-white material in its place. Therefore, some of the\n"
		                       "model may appear as bright white. To fix this, name each\n"
		                       "material in 3D Studio.");
	}

	if (doublesided) {
    logWarning(mod3dsLoad, "DOUBLE-SIDED POLYGONS\n"
		                       "This model contains double-sided polygons.\n"
		                       "The number of triangles in this model has now\n"
		                       "DOUBLED. If possible, re-do the model in\n"
		                       "3D Studio without double-sided materials.\n"
		                       "Use a normal threshold of at least 90.\n"
		                       "Also, stripification will not work properly.");
	}
//  #define INFO __FILE__ __FILE__
//  logInfo(mod3dsLoad, INFO );

	logInfo(mod3dsLoad, "File \"%s\" loaded successfully.", FileName);
	return 1;
}



H3dsScene* Loader3DS::HRead3dsScene(void)
{
	logExInfo(mod3dsLoad, "Starting file parsiing.");

	/* Initialize the global var Scene & fill with zeroes */
	Scene = (H3dsScene *) malloc(sizeof(H3dsScene));
	if (Scene == 0)
	{ logError(mod3dsLoad, "Out of memory."); return 0; }

	memset(Scene, 0, sizeof(H3dsScene));

	int retval = setjmp(EnvState);

	if(retval==0) {
		/* Return address set, start loading 3DS data. */
		word id;
		dword len, pc;

		pc=dgetpos();
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		if((int)id!=CHUNK_MAIN) {
			HFree3dsScene();
			logError(mod3dsLoad, "Error in file: No 3ds data or unknown file format.");
			return 0;
		}
		ReadMainBlocks(pc+len);
	}
	else {
		/* There was an error, free the allocated mem and return NULL. */
		HFree3dsScene();
		return 0;
	}

  logExInfo(mod3dsLoad, "File parsed successfully.");
	return Scene;
}


/* Deletes an H3dsScene from memory */
void Loader3DS::HFree3dsScene(void)
{
	int n,m;

	if(Scene != NULL)
	{
    logExInfo(mod3dsLoad, "Freeing data structures.");
		if(Scene->meshobjlist)
		{
			for(n=Scene->meshobjs-1; n>=0; n--)
			{
				H3dsMeshObj *mobj = &Scene->meshobjlist[n];
				if(mobj->maplist)
					free(mobj->maplist);
				if(mobj->facelist)
					free(mobj->facelist);
				if(mobj->vertlist)
					free(mobj->vertlist);
				if (mobj->bindings)
				{
					for (m=0; m<mobj->NumBindings; m++)
					{
						if (mobj->bindings[m].faces)
						{
							free(mobj->bindings[m].faces);
						}
					}
					free(mobj->bindings);
				}
			}
			free(Scene->meshobjlist);
		}
		if (Scene->material)
			free(Scene->material);
		free(Scene);
		Scene = NULL;
	}
}


void Loader3DS::ReadMainBlocks(dword p)
{
	uword id;
	dword len, pc;

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((unsigned int)id) {
			case CHUNK_OBJMESH:
				ReadObjMeshBlocks(pc+len);
				break;
			case CHUNK_KEYFRAMER:
				/*	ReadKeyFramerBlocks(pc+len); */
				break;
			default:
				dsetpos(pc+len);
		}
	}
}


/* Reads in mesh objects and materials */
void Loader3DS::ReadObjMeshBlocks(dword p)
{
	unsigned short id;
	dword len, pc;

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((unsigned int)id) {
			case CHUNK_MATERIAL:
				ReadMaterialBlocks(pc+len);
				break;
			case CHUNK_OBJBLOCK:
				ReadObjBlocks(pc+len);
				break;
			default: dsetpos(pc+len);
		}
	}
}


/*
	Reads in object blocks, which are cameras, meshes, or lights;
	this code only supports meshes and ignores the rest	*/
void Loader3DS::ReadObjBlocks(dword p)
{
	word id;
	dword len, pc;
	char name[MAXNAMESIZE];

	/* The object name is always the first item */
	int n=0;
	do {
		dread(&name[n++], 1);
	} while(name[n-1]!='\0' && n<sizeof(name));
	name[n-1]='\0';

	logExInfo(mod3dsLoad, "Read Object: %s.", name);

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id) {
			case CHUNK_TRIMESH: ReadTriMeshBlocks(pc+len, name); break;
		/*
			case CHUNK_LIGHT:
			case CHUNK_CAMERA:
			*/
			default: dsetpos(pc+len);
		}
	}
	dsetpos(p);
}


/*
	Eats a triangle mesh, which is a combination of XYZ vertices,
	faces, and bindings of material to faces. */
void Loader3DS::ReadTriMeshBlocks(dword p, char *name)
{
	word id;
	word fl=0; /* WG */
	dword len, pc;
	H3dsMeshObj *meshobj;
	int found_facemat;

	meshobj = EnlargeMeshArray();

	found_facemat = 0;
	strcpy(meshobj->name, name);
	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((int)id) {
			case CHUNK_VERTLIST: ReadVertList (meshobj); break;
			case CHUNK_FACELIST: fl = ReadFaceList (meshobj); break; /* WG */
			case CHUNK_MAPLIST:  ReadMapList  (meshobj); break;
			case CHUNK_TRMATRIX: ReadTraMatrix(meshobj); break;
			case CHUNK_FACEMAT:  ReadFaceMaterial(meshobj); found_facemat = 1; break;
	/* case CHUNK_SMOOLIST: */
			default: dsetpos(pc+len);
		}
	}

	if (!found_facemat) {
		defaultmaterialwarning = 1;
		ReadDefaultMaterial(meshobj, fl); /*W.G.*/
	}
}


/* Enlarge the scene array to hold a new mesh object */
H3dsMeshObj* Loader3DS::EnlargeMeshArray(void)
{
	void *mem;
	H3dsMeshObj *obj;

	mem = realloc(Scene->meshobjlist,	sizeof(H3dsMeshObj)*(Scene->meshobjs+1));
	if(mem == 0) {
		logError(mod3dsLoad, "Out of memory.");
		longjmp(EnvState, 1);
	}

	Scene->meshobjlist=(H3dsMeshObj *) mem;
	obj=&Scene->meshobjlist[Scene->meshobjs++];
	memset(obj, 0, sizeof(H3dsMeshObj));

	return obj;
}


/* Reads in the list of vertices for the current mesh object */
void Loader3DS::ReadVertList(H3dsMeshObj *meshobj)
{
	word nv;
	int n;

	dread(&nv, sizeof(nv));
	meshobj->NumVerts = nv;

	meshobj->vertlist = (H3dsVert *) getmem(sizeof(H3dsVert)*nv);
	for(n=0; n<nv; n++)
	{
		dread(&meshobj->vertlist[n].x, sizeof(float32));
		dread(&meshobj->vertlist[n].y, sizeof(float32));
		dread(&meshobj->vertlist[n].z, sizeof(float32));
	}

	logExInfo(mod3dsLoad,"Read Vertex List with %d vertices.", nv);
}


/*
	Reads in the face list for the current mesh object.
	Each face is a set of 3 indexes into the mesh object's
	vertex list and a set of options. */
word Loader3DS::ReadFaceList(H3dsMeshObj *meshobj)
{
	uword nv; // number of faces is probably not signed word, changed
	word swap;
	int n;

  dread(&nv, sizeof(nv));
  meshobj->NumFaces=nv;

	meshobj->facelist=(H3dsFace *) getmem(sizeof(H3dsFace)*nv);
  for(n=0; n<nv; n++)
  {
    dread(&meshobj->facelist[n].p0, sizeof(word));
		dread(&meshobj->facelist[n].p1, sizeof(word));
		dread(&meshobj->facelist[n].p2, sizeof(word));
		dread(&meshobj->facelist[n].flags, sizeof(word));

		/*
     * 0x07 means that the AB, BC, and CA edges are all wound
		 * in the reverse order, so swap the last two vertices
     */
    if (meshobj->facelist[n].flags == 0x07) {
			swap = meshobj->facelist[n].p1;
      meshobj->facelist[n].p1 = meshobj->facelist[n].p2;
			meshobj->facelist[n].p2 = swap;
		}
	}

  logExInfo(mod3dsLoad, "Read Face List with %d faces.", nv);

  return (nv); /* WG */
}


/* Reads in the mapping list/texture coords. for the current mesh object */
void Loader3DS::ReadMapList(H3dsMeshObj *meshobj)
{
	word nv;
	int n;

	/*  printf("Found Mapping Coordinates\n"); */

	dread(&nv, sizeof(nv));
	meshobj->NumMaps=nv;

  meshobj->maplist=(H3dsMap *) getmem(sizeof(H3dsMap)*nv);
	for(n=0; n<nv; n++)
	{
    dread(&meshobj->maplist[n].u, sizeof(float32));
		dread(&meshobj->maplist[n].v, sizeof(float32));
	}

  logExInfo(mod3dsLoad, "Read Map List with %d maps.", nv);
}


/*
	Reads in the translation matrix for the current mesh object.
	I'm not sure when this is used in a 3DS model. */
void Loader3DS::ReadTraMatrix(H3dsMeshObj *meshobj)
{
	int i;
	for (i=0;i<12;i++)
		dread(&meshobj->TraMatrix[i], sizeof(float32));
	meshobj->matrix=1;

	logExInfo(mod3dsLoad, "Read Translation Matrix:");
	logExInfo(mod3dsLoad, "%4.4f %4.4f %4.4f %4.4f", meshobj->TraMatrix[0], meshobj->TraMatrix[3], meshobj->TraMatrix[6], meshobj->TraMatrix[9]);
	logExInfo(mod3dsLoad, "%4.4f %4.4f %4.4f %4.4f", meshobj->TraMatrix[1], meshobj->TraMatrix[4], meshobj->TraMatrix[7], meshobj->TraMatrix[10]);
	logExInfo(mod3dsLoad, "%4.4f %4.4f %4.4f %4.4f.", meshobj->TraMatrix[2], meshobj->TraMatrix[5], meshobj->TraMatrix[8], meshobj->TraMatrix[11]);
}


/*
	Assigns the material in this chunk to the faces specified.
	Each material has a list of indexes into the face array to
	mark which faces use which material. */
void Loader3DS::ReadFaceMaterial(H3dsMeshObj *meshobj)
{
	uword nf;
	uword face;
	char name[50];
	H3dsMatList *binding;
	int n;

	binding = EnlargeBindingsArray(meshobj);

	/* The material name is first */
	n=0;
	do {
		dread(&name[n++], 1);
	} while(name[n-1]!='\0' && n<sizeof(name));
	name[n-1]='\0';
	strcpy(binding->name, name);

	/* Then the number of faces */
	dread(&nf, sizeof(nf));
	binding->NumFaces = (int)nf;
	binding->faces = (uword *)(malloc(sizeof(face)*nf));

	/* Bind the faces to the material */
	for (n=0; n<nf; n++) {
		dread(&face, sizeof(face));
		binding->faces[n] = face;
	}

	logExInfo(mod3dsLoad, "Read Material %s with %d faces.", name, nf);
}


void Loader3DS::ReadDefaultMaterial(H3dsMeshObj *meshobj, word nf)
{
	H3dsMatList *binding;
	int n;

	binding = EnlargeBindingsArray(meshobj);

	meshobj->NumBindings = 1;

	/* The material name is first */
	strcpy(binding->name, "WG_DEFAULT");

	/* Then the number of faces */
	binding->NumFaces = (int)nf;
	binding->faces = (uword *)(malloc(sizeof(word)*nf));

	/* Bind the faces to the material */
	for (n=0; n<nf; n++) {
		binding->faces[n] = n;
	}

	logExInfo(mod3dsLoad, "Read Material %s with %d faces.", binding->name, nf);
}


/*
	Fills out the current material with the information contained
	in these chunks */
void Loader3DS::ReadMaterialBlocks(dword p)
{
	unsigned short id;
	dword len, pc;

	H3dsMat *materialobj = EnlargeMatObjArray();

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));

		switch((unsigned int)id) {
		case CHUNK_MATNAME:
			ReadMatChunk(pc+len, materialobj);
			break;
		case CHUNK_AMBIENT:
			ReadShadingChunk(pc+len, &materialobj->ambient);
			break;
		case CHUNK_DIFFUSE:
			ReadShadingChunk(pc+len, &materialobj->diffuse);
			break;
		case CHUNK_SPECULAR:
			ReadShadingChunk(pc+len, &materialobj->specular);
			break;
		case CHUNK_TEXTURE:
			ReadTextureChunk(pc+len, &materialobj->texture);
			break;
		case CHUNK_DOUBLESIDED:
			ReadDoublesidedChunk(&materialobj->doublesided);
			break;
		case CHUNK_TRANSPARENCY:
			ReadAlphaChunk(pc+len, &materialobj->alpha);
			break;
		default:
			dsetpos(pc+len);
			break;
		}
	}
}


/* Reads in the name of the material */
void Loader3DS::ReadMatChunk(dword p, H3dsMat *mat)
{
	int n;
	char name[MAXNAMESIZE];

	n = 0;
	do {
		dread(&name[n++], 1);
	} while(name[n-1]!='\0' && n<sizeof(name));
	name[n-1]='\0';

	strcpy(mat->name, name);

	logExInfo(mod3dsLoad, "Read Material Name: %s.", name);
}


/* Reads in a color assignment */
void Loader3DS::ReadShadingChunk(dword p, H3dsRGBColor *rgb)
{
	word id;
	dword len, pc;

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((unsigned int)id) {
			case CHUNK_RGB1: ReadRGBFloat(pc+len, rgb); break;
			case CHUNK_RGB2: ReadRGBByte (pc+len, rgb); break;
			default: dsetpos(pc+len);
		}
	}
}


/* Reads in an RGB FP triplet */
void Loader3DS::ReadRGBFloat(dword p, H3dsRGBColor *rgb)
{
	dread(&rgb->red, sizeof(float32));
	dread(&rgb->green, sizeof(float32));
	dread(&rgb->blue, sizeof(float32));

	logExInfo(mod3dsLoad, "Read in RGB FP triplet:  red=%f  green=%f  blue=%f.", rgb->red, rgb->green, rgb->blue);
}


/* Reads in an RGB byte triplet */
void Loader3DS::ReadRGBByte(dword p, H3dsRGBColor *rgb)
{
	unsigned char red, green, blue;

	dread(&red, sizeof(red));
	dread(&green, sizeof(green));
	dread(&blue, sizeof(blue));
	rgb->red = (float)(red)/(float)256.0;
	rgb->green = (float)(green)/(float)256.0;
	rgb->blue = (float)(blue)/(float)256.0;

	logExInfo(mod3dsLoad, "Read in RGB byte triplet:  red=%f  green=%f  blue=%f.", rgb->red, rgb->green, rgb->blue);
}


void Loader3DS::ReadTextureChunk(dword p, H3dsTexture *texture)
{
	unsigned short id;
	dword len, pc;

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((unsigned int)id) {
			case CHUNK_MAPFILENAME:
				ReadTextureNameChunk(pc+len, texture);
				break;
			default:
				dsetpos(pc+len);
		}
	}
}


void Loader3DS::ReadTextureNameChunk(dword p, H3dsTexture *texture)
{
	int n;
	char name[MAXNAMESIZE];

	n = 0;
	do {
		dread(&name[n++], 1);
	} while(name[n-1]!='\0' && n<sizeof(name));
	name[n-1]='\0';

	strcpy(texture->texturename, name);

	logExInfo(mod3dsLoad, "Read Texture Name: %s.", name);
}


/*
 * I assume that if one material is doublesided, they all are,
 * so turn on a global doublesided flag
 */
void Loader3DS::ReadDoublesidedChunk(int *material_doublesided)
{
	*material_doublesided = 1;
	doublesided = 1;

	logExInfo(mod3dsLoad, "Doublesided Material.");
}

void Loader3DS::ReadAmountChunk(word *amount)
{
	word temp;

	dread(&temp, sizeof(temp));
	*amount = temp;
}

/*
 * I interpret the transparency information as the alpha
 * value for the entire material
 */
void Loader3DS::ReadAlphaChunk(dword p, float32 *alpha)
{
	word id;
	dword len, pc;
	word amount;

	while((pc=dgetpos()) < p) {
		dread(&id, sizeof(id));
		dread(&len, sizeof(len));
		switch((unsigned int)id) {
			case CHUNK_AMOUNT:
				ReadAmountChunk(&amount);
				break;
			default:
				dsetpos(pc+len);
		}
	}

	*alpha = (float32)amount / (float)100.0;
	if (*alpha < .01)
		*alpha = 1.0;
	else
		alphablending = 1;

	logExInfo(mod3dsLoad, "Alpha Channel: %f.", *alpha);
}



/* Retrieves the file position */
dword Loader3DS::dgetpos(void)
{
	long pos;
	if((pos=ftell(InFile)) == -1L) {
		logError(mod3dsLoad, "Error getting file position.");
		longjmp(EnvState, 1);
	}
	return (dword) pos;
}


/* Sets the file pointer to a specific position */
void Loader3DS::dsetpos(dword pos)
{
	if(fseek(InFile, (long)pos, SEEK_SET) != 0) {
		logError(mod3dsLoad, "Error moving file pointer.");
		longjmp(EnvState, 1);
	}
}


/* Reads len number of bytes from the file into the memory pointed to by *dest */
void Loader3DS::dread(void *dest, int len)
{
#ifdef WORDS_BIGENDIAN
	char *ptr, tmp1, tmp2, tmp3, tmp4;
	int x;
#endif

	if(fread(dest, len, 1, InFile) != 1) {
		logError(mod3dsLoad, "Error reading file.");
		longjmp(EnvState, 1);
	}

#ifdef WORDS_BIGENDIAN
	/* Endian Issues */
	if (len == 2) {
		/* Swap a 16-bit int */
		ptr = (char *)dest;
		tmp1 = *ptr;
		*ptr = *(ptr + 1);
		*(ptr + 1) = tmp1;
	} else if (len == 4) {
		/* Swap a 32-bit int */
		ptr = (char *)dest;
		tmp1 = *ptr;
		tmp2 = *(ptr + 1);
		tmp3 = *(ptr + 2);
		tmp4 = *(ptr + 3);
		*ptr = tmp4;
		*(ptr + 1) = tmp3;
		*(ptr + 2) = tmp2;
		*(ptr + 3) = tmp1;
	}
#endif
}


/* Safely allocates a size amount of memory */
void* Loader3DS::getmem(int size)
{
	void *mem;

	if((mem=malloc(size))==0) {
		logError(mod3dsLoad, "Out of memory (actual demand %d).", size);
		longjmp(EnvState, 1);
	}
	return mem;
}


/* Enlarge the scene array to hold a new material */
H3dsMat* Loader3DS::EnlargeMatObjArray(void)
{
	void *mem;
	H3dsMat *obj;

	mem = realloc(Scene->material, sizeof(H3dsMat)*(Scene->NumMaterials+1));
	if(mem == 0) {
		logError(mod3dsLoad, "Out of memory.");
		longjmp(EnvState, 1);
	}

	Scene->material=(H3dsMat *) mem;
	obj = &Scene->material[Scene->NumMaterials++];
	memset(obj, 0, sizeof(H3dsMat));

	return obj;
}


/* Enlarges the bindings array to hold another bindings list */
H3dsMatList* Loader3DS::EnlargeBindingsArray(H3dsMeshObj *meshobj)
{
	void *mem;
	H3dsMatList *obj;

	mem = realloc(meshobj->bindings, sizeof(H3dsMatList)*(meshobj->NumBindings+1));
	if(mem == 0) {
		logError(mod3dsLoad, "Out of memory.");
		longjmp(EnvState, 1);
	}

	meshobj->bindings = (H3dsMatList *)mem;
	obj = &meshobj->bindings[meshobj->NumBindings++];
	memset(obj, 0, sizeof(H3dsMatList));

	return obj;
}





















/* Feed our internal data to the OutGL functions */
int Loader3DS::ConvertToOutGL(void)
{
	int x,y,z;
	int current;
	float32 swap;
	RGB ambient, diffuse, specular;
	VERTEX v;
	TRIANGLE tri;
	char *name;

  logExInfo(mod3dsLoad, "Starting triange processing.");

	FreeData();
	OutGLInitialize();

	/* Start with the material properties */
	/* If the model has default materials in it, substitute an all-white material */
	if (defaultmaterialwarning)
	{
		ambient.red =
		ambient.green =
		ambient.blue = 1.0;
		ambient.alpha = 1.0;
		diffuse.red =
		diffuse.green =
		diffuse.blue = 1.0;
		diffuse.alpha = 1.0;
		specular.red =
		specular.green =
		specular.blue = 1.0;
		specular.alpha = 1.0;
		if (!OutGLAddMaterial(ambient, diffuse, specular, "WG_DEFAULT", "", 1))
			{ FreeData(); return 0; }
	}

	/* Feed each material and its properties to OutGL */
	for (x=0; x<Scene->NumMaterials; x++)
	{
		H3dsMat *mat = &Scene->material[x];
		ambient.red = mat->ambient.red;
		ambient.green = mat->ambient.green;
		ambient.blue = mat->ambient.blue;
		ambient.alpha = mat->alpha;
		diffuse.red = mat->diffuse.red;
		diffuse.green = mat->diffuse.green;
		diffuse.blue = mat->diffuse.blue;
		diffuse.alpha = mat->alpha;
		specular.red = mat->specular.red;
		specular.green = mat->specular.green;
		specular.blue = mat->specular.blue;
		specular.alpha = mat->alpha;
		if (!OutGLAddMaterial(ambient, diffuse, specular, mat->name, mat->texture.texturename, mat->doublesided))
			{ FreeData(); return 0; }
	}

	/* Then do the vertices and polygons */
	/* Order is important! */
	current = 0;
	for (x=0; x<Scene->meshobjs; x++)
	{
		H3dsMeshObj *meshobj = &Scene->meshobjlist[x];

		/* Swap y and z (.3ds wierdness) */
		for (y=0; y<meshobj->NumVerts; y++)
		{
			swap = meshobj->vertlist[y].y;
			meshobj->vertlist[y].y = meshobj->vertlist[y].z;
			meshobj->vertlist[y].z = swap;
		}

		/* Apply this material's translation matrix to this material's vertices */
		/* This doesn't seem to work as a normal transformation matrix, so this part */
		/* is commented out. */
		/*
		if (meshobj->matrix)
		{
			m = meshobj->TraMatrix;

			for (y=0; y<meshobj->NumVerts; y++)
			{
				prime.x = m[0]*meshobj->vertlist[y].x +
									m[3]*meshobj->vertlist[y].y +
									m[6]*meshobj->vertlist[y].z;
				prime.y = m[1]*meshobj->vertlist[y].x +
									m[4]*meshobj->vertlist[y].y +
									m[7]*meshobj->vertlist[y].z;
				prime.z = m[2]*meshobj->vertlist[y].x +
									m[5]*meshobj->vertlist[y].y +
									m[8]*meshobj->vertlist[y].z;

				prime.x += m[ 9];
				prime.y += m[10];
				prime.z += m[11];

				meshobj->vertlist[y].x = prime.x;
				meshobj->vertlist[y].y = prime.y;
				meshobj->vertlist[y].z = prime.z;
			}
		}
		*/

		/* Feed the vertices to outgl */
		for (y=0; y<meshobj->NumVerts; y++)
		{
			v.x = meshobj->vertlist[y].x;
			v.y = meshobj->vertlist[y].y;
			v.z = meshobj->vertlist[y].z;
			if (meshobj->NumMaps)
			{
				v.u = 1-meshobj->maplist[y].u;
				v.v = meshobj->maplist[y].v;
			} else {
				v.u = 0;
				v.v = 0;
			}
			if (!OutGLAddVertex(v))
				{ FreeData(); return 0; }
		}

		for (y=0; y<meshobj->NumBindings; y++)
		{
			name = meshobj->bindings[y].name;
			for (z=0; z<meshobj->bindings[y].NumFaces; z++)
			{
				/* Flip the polygon orientation by using points (0,2,1) */
				tri.p0 = meshobj->facelist[meshobj->bindings[y].faces[z]].p0 + current;
				tri.p1 = meshobj->facelist[meshobj->bindings[y].faces[z]].p2 + current;
				tri.p2 = meshobj->facelist[meshobj->bindings[y].faces[z]].p1 + current;
				if (!OutGLAddPolygon(tri, name))
					{ FreeData(); return 0; }
			}
		}
		current += meshobj->NumVerts;
	}

	logExInfo(mod3dsLoad, "Triangles processed.");
	HFree3dsScene();
	/* Wrap everything up in the outgl object */

	if (!OutGLFinishModel())
  	{ FreeData(); return 0; }

	FreeData();
	return 1;
}


void Loader3DS::OutGLInitialize(void)
{
	NumVertex = 0;
	NumPolyLists = 0;
	VertexList = 0;
	CommonTriangleList = 0;
	PolyList = 0;
	//center.x = center.y = center.z = 0;
	wireframe = flip = PolygonWinding = 0;
	VertexListSize = 0;
	NumPolyListsArray = 0;
	PolyListArray = NULL;
}


void Loader3DS::FreeData(void)
{
	int i;
	TRIANGLENODE *TempPtr, *LList;

	logExInfo(mod3dsLoad, "Freeing data.");

	free(VertexList);
	VertexList = NULL;
	for (i=0; i<NumPolyLists; i++)
	{
		free(PolyList[i].TriangleIndexList);
		//PolyList[i].TriangleIndexList = NULL;
		free(PolyList[i].CrunchedTriangleIndexList);
		//PolyList[i].CrunchedTriangleIndexList = NULL;
	}
	NumPolyLists = 0;
	free(PolyList);
	PolyList = NULL;

	for (i=0; i<NumVertex; i++)
	{
		LList = CommonTriangleList[i];
		while (LList)
		{
			TempPtr = LList;
			LList = LList->next;
			free (TempPtr);
		}
	}
	NumVertex = 0;
	free(CommonTriangleList);
	CommonTriangleList = NULL;

	for (i = 0; i< NumPolyListsArray; i++)
		free(PolyListArray[i].array);
	NumPolyListsArray = 0;
	free(PolyListArray);
	PolyListArray = NULL;
}


/* Add a material to the model */
int Loader3DS::OutGLAddMaterial(RGB ambient, RGB diffuse, RGB specular, char *name, char *texname, int doublesided)
{
	void *mem;
	int r;

	if ((mem = realloc(PolyList, sizeof(POLYGONLIST) * (NumPolyLists+1))) == 0)
	{
		logError(mod3dsLoad, "Out of memory.");
		return 0;
	}

	PolyList = (POLYGONLIST *)mem;

#ifdef N_LOAD_TEXTURES
	PolyList[NumPolyLists].textureID = 0;
#else
	if (strlen(texname)>0)
	{
        /* create texture */
        Texture *T = new(nothrow) Texture;
        if (T==NULL)
        {
            PolyList[NumPolyLists].textureID = 0;
            logError(mod3dsLoad, "Can't create texture object - Out of memory - material %d, texture %s.", NumPolyLists, &texname);
            r = 0;
        }
        else
        {
			int L1 = strlen(ScenePath);
            char *FN = (char*)malloc(L1+strlen(texname)+1);
			strcpy(FN, &ScenePath[0]);
            strcpy(FN+L1, &texname[0]);
			if (T->LoadTGA(FN))
/*			unsigned char *P = new unsigned char[256*256*4];
			for(int i = 0; i<256; i++)
				for(int j=0; j<256; j++)
				{
					P[i*1024 + j*4 + 0] = j;
					P[i*1024 + j*4 + 1] = 0;
					P[i*1024 + j*4 + 2] = 255;
					P[i*1024 + j*4 + 3] = 32;
				}
			if (T->LoadFromMemory(P,256,256))*/
			{
				PolyList[NumPolyLists].textureID = T->TexName();
				logExInfo(mod3dsLoad, "Material %d, texture %s, cislo texury %d.", NumPolyLists, &texname, PolyList[NumPolyLists].textureID);
				r = 1;
            }
            else
            {
                PolyList[NumPolyLists].textureID = 0;
                r = 0;
            }
            free(FN);
        }
	}
	else
	{
		PolyList[NumPolyLists].textureID = 0;
		logExInfo(mod3dsLoad, "Material %d, cislo texury 0.", NumPolyLists);
		r = 1;
	}
#endif

	strcpy(PolyList[NumPolyLists].name, name);
	strcpy(PolyList[NumPolyLists].texturename, texname);
	PolyList[NumPolyLists].ambient = ambient;
	PolyList[NumPolyLists].diffuse = diffuse;
	PolyList[NumPolyLists].specular = specular;
	PolyList[NumPolyLists].doublesided = doublesided;
	PolyList[NumPolyLists].NumTriangles = 0;
	PolyList[NumPolyLists].TriangleIndexList = NULL;
	PolyList[NumPolyLists].CrunchedTriangleIndexList = NULL;

	NumPolyLists++;
	return r;
}


/* Adds a vertex to the vertex list */
int Loader3DS::OutGLAddVertex(VERTEX v)
{
	void *mem;

	if (NumVertex == VertexListSize)
	{
		VertexListSize += VERTEXBLOCKSIZE;
		if ((mem = realloc(VertexList, sizeof(VERTEX) * (VertexListSize))) == 0)
		{
			logError(mod3dsLoad, "Out of memory.");
			return 0;
		}
		VertexList = (VERTEX *)mem;

		if ((mem = realloc(CommonTriangleList, sizeof(TRIANGLENODE **) * (VertexListSize))) == 0)
		{
			logError(mod3dsLoad, "Out of memory.");
			return 0;
		}
		CommonTriangleList = (TRIANGLENODE **)mem;
	}

	VertexList[NumVertex] = v;

	/*center.x += v.x;
	center.y += v.y;
	center.z += v.z;*/

	/*
	 * Add a slot into the array we're building to keep track of which
	 * triangles are part of which vertices, which will be used to build
	 * vertex normals.
	 */
	CommonTriangleList[NumVertex] = 0;

	NumVertex++;
	return 1;
}


/* Add a polygon of the material */
int Loader3DS::OutGLAddPolygon(TRIANGLE tri, char *MaterialName)
{
	void *mem;
	int MatNum;
	POLYGONLIST *list;
	SURFACE TempSurface;

	TempSurface.p0 = tri.p0;
	TempSurface.p1 = tri.p1;
	TempSurface.p2 = tri.p2;

	/* Ignore degenerate polygons */
	if (CalculateSurfaceNormal(&TempSurface) == DEGENERATE_SURFACE)
		return 1;

	/* Find the polygon list index we belong to */
	MatNum = FindMaterialIndex(MaterialName);
	if (MatNum == -1)
		return 0;
	list = &PolyList[MatNum];

	/* Attach a new triangle to the polygon list */
	if ((mem = realloc(list->TriangleIndexList,
				 sizeof(SURFACE) * (list->NumTriangles+1))) == 0)
	{
		logError(mod3dsLoad, "Out of memory.");
		return 0;
	}
	list->TriangleIndexList = (SURFACE *)mem;
	list->TriangleIndexList[list->NumTriangles].p0 = tri.p0;
	list->TriangleIndexList[list->NumTriangles].p1 = tri.p1;
	list->TriangleIndexList[list->NumTriangles].p2 = tri.p2;
	CalculateSurfaceNormal(&list->TriangleIndexList[list->NumTriangles]);
	list->NumTriangles++;

	if (PolyList[MatNum].doublesided)
	{
		/* Attach a new triangle to the polygon list */
		if ((mem = realloc(list->TriangleIndexList,
					 sizeof(SURFACE) * (list->NumTriangles+1))) == 0)
		{
			logError(mod3dsLoad, "Out of memory.");
			return 0;
		}
		list->TriangleIndexList = (SURFACE *)mem;
		list->TriangleIndexList[list->NumTriangles].p0 = tri.p0;
		list->TriangleIndexList[list->NumTriangles].p1 = tri.p2;
		list->TriangleIndexList[list->NumTriangles].p2 = tri.p1;
		CalculateSurfaceNormal(&list->TriangleIndexList[list->NumTriangles]);
		list->NumTriangles++;
	}
	return 1;
}


/* Finds the index of the material in question */
int Loader3DS::FindMaterialIndex(char *name)
{
	int n;

	for (n=0; n<NumPolyLists; n++)
	{
		if (strcmp(name, PolyList[n].name) == 0)
		{
			return n;
		}
	}

	return -1;
}


int Loader3DS::CalculateSurfaceNormal(SURFACE *ThisTriangle)
{
	NORMAL ThisNormal;
	GLfloat Ai, Aj, Ak;
	GLfloat Bi, Bj, Bk;
	GLfloat sum, magnitude;

	/*
		Calculate the polygon's normal using a cross product.
		A,B are vectors; i,j,k are their xyz components
		*/
	Ai = VertexList[ThisTriangle->p1].x - VertexList[ThisTriangle->p0].x;
	Aj = VertexList[ThisTriangle->p1].y - VertexList[ThisTriangle->p0].y;
	Ak = VertexList[ThisTriangle->p1].z - VertexList[ThisTriangle->p0].z;
	Bi = VertexList[ThisTriangle->p2].x - VertexList[ThisTriangle->p1].x;
	Bj = VertexList[ThisTriangle->p2].y - VertexList[ThisTriangle->p1].y;
	Bk = VertexList[ThisTriangle->p2].z - VertexList[ThisTriangle->p1].z;

	/* Find the cross product */
	ThisNormal.n0 = Aj*Bk - Ak*Bj;
	ThisNormal.n1 = Ak*Bi - Ai*Bk;
	ThisNormal.n2 = Ai*Bj - Aj*Bi;

	/* Make the normal a unit vector */
	sum = (ThisNormal.n0 * ThisNormal.n0) +
		(ThisNormal.n1 * ThisNormal.n1) +
		(ThisNormal.n2 * ThisNormal.n2);

	/*
	 * This will be zero only if the cross product's
	 * normal has all zero components. That won't happen
	 * as long as the triangle has three separate vertices.
	 * Otherwise, the next step would be to divide 1 by the
	 * square root of zero... not a pretty site.
	 */
	if (sum == 0)
		return DEGENERATE_SURFACE;

	magnitude = 1/(GLfloat)sqrt(sum);

	ThisNormal.n0 *= magnitude;
	ThisNormal.n1 *= magnitude;
	ThisNormal.n2 *= magnitude;

	ThisTriangle->n0 = ThisNormal.n0;
	ThisTriangle->n1 = ThisNormal.n1;
	ThisTriangle->n2 = ThisNormal.n2;

	return JOB_DONE;
}



























/*
 * Performs some model clean-up that can only be done once we have the entire
 * model. This includes finding the vertex normals (instead of the polygon
 * normals we already calculated), centering the object around 0,0,0, crunching
 * the vertex data into vertex arrays for OpenGL, and compiling the display list.
 * Note that this code finds the vertex normals by first calculating the angle
 * that each polygon shares with that vertex. It then averages all the normals
 * together, each one weighted on how much of an angle they contribute to the
 * vertex. Also, if the user specified a threshold for normals so that surface
 * edges are preserved, we have to add a new vertex to the list with the same
 * X,Y,Z coordinates but different normal coordinates. Thus, edge preservation
 * may look good, but it's expensive.
 */
int Loader3DS::OutGLFinishModel(void)
{
	int OriginalNumVertex;
	int NumPolygons;
	int i;

	logExInfo(mod3dsLoad, "Starting completion of model.");

	OriginalNumVertex = NumVertex;

	logExInfo(mod3dsLoad, "Total number of vertices in model: %d.", NumVertex);

	LinkVerticesToPolygons();
	if (!CalculateVertexNormals())
    	return 0;
	MakeTriangleIndexList();

	if (UseThreshold)
		logExInfo(mod3dsLoad, "Generated %d new vertex/normals for edge preservation.", NumVertex - OriginalNumVertex);
	logExInfo(mod3dsLoad, "Total number of vertices: %d.", NumVertex);

	NumPolygons = 0;
	for (i=0; i<NumPolyLists; i++)
		NumPolygons += PolyList[i].NumTriangles;
	logInfo(mod3dsLoad, "Total number of polygons: %d.", NumPolygons);

	/* seazen polygon list podle textur */
	if (!TextureIndex())
		return 0;

	logExInfo(mod3dsLoad, "Completion of model finished. Drawing...");

	return OutGLCompileDisplayList();
}


/*
 * Build links from the vertices back to the polygon lists
 */
void Loader3DS::LinkVerticesToPolygons(void)
{
	int i,j;
	SURFACE *ThisTriangle;
	TRIANGLENODE *TempPtr;

	for (i=0; i<NumPolyLists; i++)
	{
		for (j=0; j<PolyList[i].NumTriangles; j++)
		{

			/* For each triangle, insert a pointer to the triangle */
			/* into a linked list hanging off of each of the three */
			/* vertices it touches.                                */
			ThisTriangle = &PolyList[i].TriangleIndexList[j];

			/* Point One */
			TempPtr = (TRIANGLENODE *) malloc(sizeof(TRIANGLENODE));
			TempPtr->triangle = ThisTriangle;
			TempPtr->next = CommonTriangleList[ThisTriangle->p0];
			CommonTriangleList[ThisTriangle->p0] = TempPtr;

			/* Point Two */
			TempPtr = (TRIANGLENODE *) malloc(sizeof(TRIANGLENODE));
			TempPtr->triangle = ThisTriangle;
			TempPtr->next = CommonTriangleList[ThisTriangle->p1];
			CommonTriangleList[ThisTriangle->p1] = TempPtr;

			/* Point Three */
			TempPtr = (TRIANGLENODE *) malloc(sizeof(TRIANGLENODE));
			TempPtr->triangle = ThisTriangle;
			TempPtr->next = CommonTriangleList[ThisTriangle->p2];
			CommonTriangleList[ThisTriangle->p2] = TempPtr;
		}
	}

	/* Sample code that traverses the list we just made above */
	/*
	j=0;
  for (i=0; i<NumVertex; i++) {
    for (TempPtr = CommonTriangleList[i]; TempPtr; TempPtr=TempPtr->next) {
	j++;
	}
	}
  */
}

/*
 * Generate the per-vertex normals. Make new vertices if two surfaces are
 * more than a threshold angle apart. Thus, a cube would have six surface
 * normals but if the threshold was 90 degrees, it would have 24
 * vertex/normal pairs.
 */
int Loader3DS::CalculateVertexNormals(void)
{
//	void *mem;
	int i;
//	int NewVertexAlreadyMade;
//	int NewVertex;
//	int NumOriginalVertex;
	GLfloat VertexAngle;
	GLfloat TotalAngles;
	TRIANGLENODE *ThisSurface;
//	TRIANGLENODE *TempPtr;
	GLfloat sum, magnitude;

//	NumOriginalVertex = NumVertex;
//	NewVertex = 0;

	/* Calculate a normal for every vertex */
	for (i=0; i<NumVertex; i++)
	{
		VertexList[i].nx = VertexList[i].ny = VertexList[i].nz = 0;
		TotalAngles = 0;
//		NewVertexAlreadyMade = 0;
		ThisSurface = CommonTriangleList[i];
//		TotalAngles = 0;

		/* Traverse the linked list of triangles that touch this vertex */
		while (ThisSurface)
		{
			#ifndef DISABLE_THRESHOLD
			if (UseThreshold)
      {
        GLuint WantedVertex;
        TRIANGLENODE *BrowsingSurface;
        if (ThisSurface->triangle->p0 != i)
	        WantedVertex = ThisSurface->triangle->p0;

        while (BrowsingSurface)
		{
          if (WantedVertex == BrowsingSurface->triangle->p0)
          {
        		if (cos_angle(ThisSurface->triangle, BrowsingSurface->triangle) <= cos_threshold)
            {

            }

		  }

        }


			{
				/* Put this triangle at another vertex */
				if (!NewVertexAlreadyMade)
				{
					/* Make a copy of this vertex at the end of the vertex list */
					NewVertex++;
					NewVertexAlreadyMade = 1;
					if ((mem = realloc(VertexList, sizeof(VERTEX) * (NumVertex+1))) == 0)
					{
						logError(mod3dsLoad, "Error reallocating memory.");
						return 0;
					}
					VertexList = (VERTEX *)mem;
					VertexList[NumVertex].x = VertexList[i].x;
					VertexList[NumVertex].y = VertexList[i].y;
					VertexList[NumVertex].z = VertexList[i].z;
					VertexList[NumVertex].u = VertexList[i].u;
					VertexList[NumVertex].v = VertexList[i].v;

					if ((mem = realloc(CommonTriangleList, sizeof(TRIANGLENODE) * (NumVertex+1))) == 0)
					{
						logError(mod3dsLoad, "Error reallocating memory.");
						return 0;
					}
					CommonTriangleList = (TRIANGLENODE **)mem;
					CommonTriangleList[NumVertex] = 0;

					NumVertex++;
				}
				/*
				 * Attach this surface to the last vertex's list of adjacent triangles
				 * Must be in this order or our ->next's won't link correctly
				 */
				TempPtr = ThisSurface;
				ThisSurface = ThisSurface->next;
				TempPtr->next = CommonTriangleList[NumVertex-1];
				CommonTriangleList[NumVertex-1] = TempPtr;

				/*
				 * Change the vertex indices so that we point to the new vertex
				 * instead of the one we were originally pointing to.
				 */
				if ((signed)TempPtr->triangle->p0 == i) {
					TempPtr->triangle->p0 = NumVertex-1;
				}
				if ((signed)TempPtr->triangle->p1 == i) {
					TempPtr->triangle->p1 = NumVertex-1;
				}
				if ((signed)TempPtr->triangle->p2 == i) {
					TempPtr->triangle->p2 = NumVertex-1;
				}
			} else
      #endif
      {
				/*
				 * Since it's less then the threshold angle, calculate the angle that this
				 * polygon shares with the vertex
				 */
				VertexAngle = angle(ThisSurface->triangle, i);
				VertexList[i].nx += ThisSurface->triangle->n0 * VertexAngle;
				VertexList[i].ny += ThisSurface->triangle->n1 * VertexAngle;
				VertexList[i].nz += ThisSurface->triangle->n2 * VertexAngle;
				TotalAngles += VertexAngle;

				ThisSurface = ThisSurface->next;
			}
		}
		if (TotalAngles>1)
		{
			/* Calculate the normal for this vertex */
			VertexList[i].nx /= TotalAngles;
			VertexList[i].ny /= TotalAngles;
			VertexList[i].nz /= TotalAngles;

			/* Make it a unit vector again */
			sum = ((VertexList[i].nx * VertexList[i].nx) +
			 (VertexList[i].ny * VertexList[i].ny) +
			 (VertexList[i].nz * VertexList[i].nz));
			if (sum != 0) /* if (sum==0)  keep normal length zero */
            {
				magnitude = 1/(GLfloat)sqrt(sum);

				VertexList[i].nx *= magnitude;
				VertexList[i].ny *= magnitude;
				VertexList[i].nz *= magnitude;
            }
		}
	}

	//if (centerModel) {

		/* Adjust the model to center around (0,0,0) */
/*
		center.x /= NumOriginalVertex;
		center.y /= NumOriginalVertex;
		center.z /= NumOriginalVertex;

		for (i=0; i<NumVertex; i++) {
			VertexList[i].x -= center.x;
			VertexList[i].y -= center.y;
			VertexList[i].z -= center.z;
		}*/
//	}
  	return 1;
}


/*
 * Make a list of triangle indexes for OpenGL's vertex arrays
 * This does not produce the more optimized TRIANGLE_STRIPS
 */
void Loader3DS::MakeTriangleIndexList(void)
{
	int i,j;

	for (i=0; i<NumPolyLists; i++)
  	{
    	PolyList[i].CrunchedTriangleIndexList = (TRIANGLE*)malloc(sizeof(TRIANGLE) * (PolyList[i].NumTriangles+1));
    	for (j=0; j<PolyList[i].NumTriangles; j++)
    	{
      		PolyList[i].CrunchedTriangleIndexList[j].p0 = PolyList[i].TriangleIndexList[j].p0;
			PolyList[i].CrunchedTriangleIndexList[j].p1 = PolyList[i].TriangleIndexList[j].p1;
      		PolyList[i].CrunchedTriangleIndexList[j].p2 = PolyList[i].TriangleIndexList[j].p2;
		}
  	}
}


/* vyhledvac funkce pro pole textur */
int Loader3DS::FindTextureIndex(GLuint tex)
{
	int i;
	for (i = 0; i< NumPolyListsArray; i++)
		if (PolyList[PolyListArray[i].array[0]].textureID == tex)
			return i;
	return -1;
}

/* vytvoen pole seznam objekt se stejnou texturou */
int Loader3DS::TextureIndex(void)
{
	void *mem;
	int i, index;

	for (i = 0; i <NumPolyLists; i++)
	{
		if ((index = FindTextureIndex(PolyList[i].textureID)) == -1)
		{
			if ((mem = realloc(PolyListArray,
            	sizeof(SURFACE) * (NumPolyListsArray+1))) == NULL)
			{
				logError(mod3dsLoad, "Error reallocating memory.");
				return 0;
            }
			PolyListArray = (POLYGONLISTARRAY *) mem;

			if ((mem = malloc(sizeof(int))) == NULL)
			{
				logError(mod3dsLoad, "Error reallocating memory.");
            	return 0;
			}
			PolyListArray[NumPolyListsArray].array = (int *) mem;
			PolyListArray[NumPolyListsArray].NumPolyLists = 1;
			PolyListArray[NumPolyListsArray].array[0] = i;
			NumPolyListsArray++;
		}
		else
		{
			if ((mem = realloc(PolyListArray[index].array,
            	sizeof(int) * (PolyListArray[index].NumPolyLists + 1))) == NULL)
			{
				logError(mod3dsLoad, "Error reallocating memory.");
				return 0;
		 	}
			PolyListArray[index].array = (int *) mem;
			PolyListArray[index].array[PolyListArray[index].NumPolyLists] = i;
			PolyListArray[index].NumPolyLists++;
		}
	}
	#ifdef DEBUG_3DS
    for (i = 0; i <NumPolyListsArray; i++)
	{
		int j;
		logExInfo(mod3dsLoad, "pocet objektu v %d seznamu: %d, cislo textury: %d, objekty:",
			i, PolyListArray[i].NumPolyLists, PolyList[PolyListArray[i].array[0]].textureID);
		for (j = 0; j <PolyListArray[i].NumPolyLists; j++)
			logExInfo(mod3dsLoad, " %d", PolyListArray[i].array[j]);
	}
#endif
	return 1;
}


/* Compiles the display list with the current options */
int Loader3DS::OutGLCompileDisplayList()
{
	int i, j;
    bool DLCreated = (DisplayList <= 0);

    if (DLCreated)
    	DisplayList = glGenLists(1);

	glInterleavedArrays(GL_T2F_N3F_V3F, 0, VertexList);

	glNewList(DisplayList, GL_COMPILE);

	assert(NumPolyListsArray>0);
	int tn;
	int tl = 0;
	if (PolyList[PolyListArray[0].array[0]].textureID == 0)
		glDisable(GL_TEXTURE_2D);

	for (i=0; i<NumPolyListsArray; i++) {

		tn = PolyList[PolyListArray[i].array[0]].textureID;
		if (tn != tl)
		{
			if (tn > 0) {
				if (tl == 0)
					glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tn);
			} else
				glDisable(GL_TEXTURE_2D);

			tl = tn;
		}

		for (j = 0; j < PolyListArray[i].NumPolyLists; j++)
		{
			glMaterialfv(GL_FRONT, GL_AMBIENT,
							 (GLfloat *) &PolyList[PolyListArray[i].array[j]].ambient);
			glMaterialfv(GL_FRONT, GL_DIFFUSE,
							 (GLfloat *) &PolyList[PolyListArray[i].array[j]].diffuse);
			glMaterialfv(GL_FRONT, GL_SPECULAR,
							 (GLfloat *) &PolyList[PolyListArray[i].array[j]].specular);
			glColor3fv((GLfloat *) &PolyList[PolyListArray[i].array[j]].diffuse);

			glDrawElements(GL_TRIANGLES,
					 PolyList[PolyListArray[i].array[j]].NumTriangles*3,
					 GL_UNSIGNED_INT,
					 &PolyList[PolyListArray[i].array[j]].CrunchedTriangleIndexList[0]);
		}
	}
	glEndList();

	if (glGetError() != GL_NO_ERROR) {
    	if (DLCreated)
        	glDeleteLists(DisplayList,1);
		logError(mod3dsLoad, "Error compiling display list.");
		return 0;
	}
	return 1;
}


/*
 * Returns the cosine of the angle between two normals.
 * Since the normals are already normalized, we don't need to
 * divide by the product of their magnitudes. We only need
 * to return the dot product since the denominator will always
 * be 1.
 */
GLfloat Loader3DS::cos_angle(SURFACE *n1, SURFACE *n2) {
	GLfloat dot;

	dot = n1->n0 * n2->n0 +
	n1->n1 * n2->n1 +
	n1->n2 * n2->n2;

	return dot;
}

/*
 * Returns the angle of a triangle at the specified
 * vertex.
 * square(c) = square(a) + square(b) + 2*a*b*cos(theta)
 * which is a general form of Pythagorean's theorem.
 */
GLfloat Loader3DS::angle(SURFACE *triangle, GLuint vertex)
{
	double result;
	double a, b, c;

  if (triangle->p0 == vertex) {
    a = distance(&VertexList[triangle->p0], &VertexList[triangle->p1]);
    b = distance(&VertexList[triangle->p0], &VertexList[triangle->p2]);
    c = distance(&VertexList[triangle->p1], &VertexList[triangle->p2]);
	} else if (triangle->p1 == vertex) {
		a = distance(&VertexList[triangle->p1], &VertexList[triangle->p0]);
    b = distance(&VertexList[triangle->p1], &VertexList[triangle->p2]);
		c = distance(&VertexList[triangle->p2], &VertexList[triangle->p0]);
  } else if (triangle->p2 == vertex) {
    a = distance(&VertexList[triangle->p2], &VertexList[triangle->p0]);
    b = distance(&VertexList[triangle->p2], &VertexList[triangle->p1]);
		c = distance(&VertexList[triangle->p0], &VertexList[triangle->p1]);
  } else {
    return 0;
  }

	result = c*c - a*a - b*b;
	result /= 2*a*b;
	if (fabs(result) >= 1.0) // sometimes needed
    {
	    if (result>=1.0)
    		result = 1.0;
	    else
    		if (result<=-1.0)
        		result = -1.0;
    }
	result = acos(result);

	return (GLfloat) result;
}


/* Distance between two 3D points */
double Loader3DS::distance(VERTEX *v1, VERTEX *v2)
{
	double dx, dy, dz;
	double result;

	dx = v1->x - v2->x;
	dy = v1->y - v2->y;
	dz = v1->z - v2->z;

	result = sqrt(dx*dx + dy*dy + dz*dz);

	return result;
}









static Loader3DS *Loader;


int Load3DSInit(void)
{
#ifdef USEMOD_3DSLOAD
	Loader = new(nothrow) Loader3DS;
	if (Loader == NULL) return 0;
#endif
	return 1;
}


void Load3DSDone(void)
{
#ifdef USEMOD_3DSLOAD
	delete Loader;
#endif
}

int Load3DS(char* FileName)
{
#ifdef USEMOD_3DSLOAD
	int r;
	int L1 = strlen(ScenePath);
	char *FN = (char*)malloc(L1+strlen(FileName)+1);
	strcpy(FN, &ScenePath[0]);
	strcpy(FN+L1, &FileName[0]);
	Loader->DisplayList = 0;
	r = Loader->Load3ds(FN);
	free(FN);
	if (r) return Loader->DisplayList;
	else return 0;
#else
	return 1;
#endif
}









