/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */


#include <list>
#define WANT_MEMSET
#include "porting.h"

#include "Timers.h"
#include "Log.h"
#include "InputDevices.h"
#include "Input.h"

#include "setup.h"

#ifdef USE_SDL
# include <SDL/SDL_keysym.h>
#endif
#ifdef USE_GLUT
# define GLUT_API_VERSION 4
# define GLUT_BUILDING_LIB
# include <GL/glut.h>
#endif


InputAxesTable InputAxes;
InputKeysTable InputKeys;
static void KeysUpdate();
static void LoadAssignmentsDefaults(void);



timType LastUpdateTime = 0;
float MouseSensitivityHorizontal = (float)1.0/50;  // turn (in 1/rad) at 1 second when mouse is moved 10 points right
float MouseSensitivityVertical = (float)1.0/50;  // turn (in 1/rad) at 1 second when mouse is moved 10 points up
float KeyboardAssignedSpeed = 10;



class KeyAssignment
{
public:
	int Key;
	bool *B;
	void Zero()  { (*B) = false; }
	virtual void Update() = 0;
};
class KeyAssignmentDown : public KeyAssignment
{
public:
	virtual void Update()  { (*B) = (*B) || (KeyboardKeyState[Key].Down || KeyboardKeyState[Key].Pressed); };
};
class KeyAssignmentPress : public KeyAssignment
{
public:
	virtual void Update()  { (*B) = (*B) || KeyboardKeyState[Key].Pressed; };
};
std::list <KeyAssignmentDown> KeyAssignmentsDown;
std::list <KeyAssignmentPress> KeyAssignmentsPress;

static struct {
	bool RotL;    // rotation left-right - axes y
	bool RotR;    // rotation left-right - axes y
	bool RotDU;    // rotation down-up - axes x
	bool Roll;     // rotation rolling - axes z
	bool MovForward;
	bool MovBackward;
	bool MovL;
	bool MovR;
	bool MovD;
	bool MovU;
} AxesByKeys;


static timType LastTickTime;
int InputUpdate(void)
{
#ifdef USEMOD_INPUT
	int r;
	int x,y;
# define dt frLastFrameTime

	logInfo(modInput, "Getting status from input devices");

	r = MouseGetData(x,y);

	if (r)
	{
		InputAxes.RotLR = (float)x * MouseSensitivityHorizontal;
		InputAxes.RotDU = (float)y * MouseSensitivityVertical;
	} else {
		InputAxes.RotLR = 0;
		InputAxes.RotDU = 0;
		logWarning(modInput, "Can't read mouse state.");
	}


	if (!KeyboardUpdate())
	{
		r = 0;
		InputAxes.MovForward = 0;
		InputAxes.MovLR = 0;
//		memset(&InputState, 0, sizeof(InputState));
		logWarning(modInput, "Can't read keyboard state.");
	}
	else
	{
		KeysUpdate();

		if (AxesByKeys.MovL)
			if (AxesByKeys.MovR)
				InputAxes.MovLR = 0;
			else
				InputAxes.MovLR = -KeyboardAssignedSpeed * (float)frLastFrameTime;
		else
			if (AxesByKeys.MovR)
				InputAxes.MovLR = +KeyboardAssignedSpeed * (float)frLastFrameTime;
			else
				InputAxes.MovLR = 0;

		if (AxesByKeys.MovD)
			if (AxesByKeys.MovU)
				InputAxes.MovDU = 0;
			else
				InputAxes.MovDU = -KeyboardAssignedSpeed * (float)frLastFrameTime;
		else
			if (AxesByKeys.MovU)
				InputAxes.MovDU = +KeyboardAssignedSpeed * (float)frLastFrameTime;
			else
				InputAxes.MovDU = 0;

		if (AxesByKeys.MovBackward)
			if (AxesByKeys.MovForward)
				InputAxes.MovForward = 0;
			else
				InputAxes.MovForward = -KeyboardAssignedSpeed * (float)frLastFrameTime;
		else
			if (AxesByKeys.MovForward)
				InputAxes.MovForward = +KeyboardAssignedSpeed * (float)frLastFrameTime;
			else
				InputAxes.MovForward = 0;
	}
	KeyboardClearPressed();


	return r;
#else
	return 1;
#endif
}


int InputInit(void)
{
#ifdef USEMOD_INPUT
	int r = 1;
	logInfo(modInput, "Starting devices initialization.");

	memset(&InputAxes, 0, sizeof(InputAxes));
	LoadAssignmentsDefaults();

	if (!MouseInit())
	{
		logError(modInput, "Can't initialize mouse.");
		r = 0;
	}
	if (!KeyboardInit())
	{
		logError(modInput, "Can't initialize keyboard.");
		r = 0;
	}

	logInfo(modInput, "Device initialization finished.");
	return r;
#else
	return 1;
#endif
}


void InputDone(void)
{
#ifdef USEMOD_INPUT
	logInfo(modInput, "Starting devices shooting down.");
	MouseDone();
	KeyboardDone();
	logInfo(modInput, "All devices terminated successfully.");
#endif
}


void InputAcquire(void)
{
#ifdef USEMOD_INPUT
	logInfo(modInput, "Acquiring devices.");
	if (!KeyboardAcquire())
		logWarning(modInput, "Can't acquire keyboard device.");
	if (!MouseAcquire())
		logWarning(modInput, "Can't acquire mouse device.");
	LastTickTime = timGetTicks();
#endif
}


void InputUnacquire(void)
{
#ifdef USEMOD_INPUT
	logInfo(modInput, "Unacquiring devices.");
	KeyboardUnacquire();
	MouseUnacquire();
#endif
}


void InputWindowCallback(void)
{
#ifdef USEMOD_INPUT
	logInfo(modInput, "Window modified callback.");
	if (!MouseWindowCallback())
		logError(modInput, "Can't reacquire mouse to modified window.");
#endif
}


void InputAssignKeyDown(const int Ch, bool *b)
{
	KeyAssignmentDown n;
	n.Key = Ch;
	n.B = b;
	KeyAssignmentsDown.push_back(n);
}

void InputAssignKeyPress(const int Ch, bool *b)
{
	KeyAssignmentPress n;
	n.Key = Ch;
	n.B = b;
	KeyAssignmentsPress.push_back(n);
}


static void KeysUpdate()
{
	std::list<KeyAssignmentPress>::iterator listIt1;
	std::list<KeyAssignmentDown>::iterator listIt2;

	for (listIt1 = KeyAssignmentsPress.begin(); listIt1 != KeyAssignmentsPress.end(); listIt1++)
		listIt1->Zero();
	for (listIt2 = KeyAssignmentsDown.begin(); listIt2 != KeyAssignmentsDown.end(); listIt2++)
		listIt2->Zero();

	for (listIt1 = KeyAssignmentsPress.begin(); listIt1 != KeyAssignmentsPress.end(); listIt1++)
		listIt1->Update();
	for (listIt2 = KeyAssignmentsDown.begin(); listIt2 != KeyAssignmentsDown.end(); listIt2++)
		listIt2->Update();
}
 

static void LoadAssignmentsDefaults(void)
{
#ifdef USE_SDL
	InputAssignKeyDown(SDLK_e,   &AxesByKeys.MovForward);
	InputAssignKeyDown(SDLK_d,   &AxesByKeys.MovBackward);
	InputAssignKeyDown(SDLK_f,   &AxesByKeys.MovR);
	InputAssignKeyDown(SDLK_s,   &AxesByKeys.MovL);
	InputAssignKeyDown(SDLK_r,   &AxesByKeys.MovU);
	InputAssignKeyDown(SDLK_x,   &AxesByKeys.MovD);
	InputAssignKeyDown(SDLK_UP,    &AxesByKeys.MovForward);
	InputAssignKeyDown(SDLK_DOWN,  &AxesByKeys.MovBackward);
	InputAssignKeyDown(SDLK_RIGHT, &AxesByKeys.MovR);
	InputAssignKeyDown(SDLK_LEFT,  &AxesByKeys.MovL);
	InputAssignKeyPress(SDLK_SPACE,&InputKeys.Fire);

	InputAssignKeyPress(SDLK_F5,    &InputKeys.ChangeCamera);

	InputAssignKeyPress(SDLK_F1,    &InputKeys.FPSKey);
	InputAssignKeyPress(SDLK_F2,    &InputKeys.Pause);
	InputAssignKeyPress(SDLK_ESCAPE,&InputKeys.Terminate);
#endif
#ifdef USE_GLUT
	InputAssignments.MovF.AssignKeys(GLUT_KEY_UP,0,0,0);
	InputAssignments.MovB.AssignKeys(GLUT_KEY_DOWN,0,0,0);
	InputAssignments.MovR.AssignKeys(GLUT_KEY_RIGHT,0,0,0);
	InputAssignments.MovL.AssignKeys(GLUT_KEY_LEFT,0,0,0);
	InputAssignments.ChangeCamera.AssignKeys(GLUT_SPACE,0,0,0);
	InputAssignments.FPSKey.AssignKeys(GLUT_F1,0,0,0);
#endif
}



