/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */


#include "config.h"
#ifdef WINDOZE
#include <windows.h>
#endif
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>
#include <limits.h>
#pragma hdrstop // for C++ Builder compiler speed up

#include "Log.h"
#define WANT_VSNPRINTF
#include "porting.h"

#include "setup.h"




//#define LOG_SHOW_ALWAYS
#define LOG_SHOW_ON_WARNINGS
//#define LOG_SHOW_ON_ERRORS
// what about fatal errors?

#define LOG_FLUSH_ALWAYS
//#define LOG_FLUSH_WARNINGS
//#define LOG_FLUSH_ERRORS



logLevel logLevelTable[modUnknown+1] =
{
	logLevelExInfos, // Errors
	logLevelExInfos, // Main
	logLevelInfos, // GUI
	logLevelExInfos, // Timers
	logLevelExInfos, // Fonts
	logLevelExInfos, // Cockpit
	logLevelWarnings, // Input
	logLevelInfos, // 3dsLoad
	logLevelExInfos, // Scene
	logLevelExInfos, // SceneClasses
	logLevelExInfos, // Textures
	logLevelExInfos, // TgaPcxLoad
	logLevelExInfos, // AI
	logLevelExInfos  // Unknown
};

static char LOG_FILE_NAME[] = "VRapps.log";
static char LOG_FILE_CREATE_ERROR[] = "Can't create log file (VRapps.log).\nCheck access rights and disk free space.";
static char LOG_FILE_CLOSE_ERROR[] = "Can't close log file (VRapps.log).\nCheck access rights and disk free space.";
static char LOG_FILE_READ_ERROR[] = "Can't read log file (VRapps.log).\nCheck access rights and disk free space.";
static char LOG_FILE_WINDOW_ERROR[] = "Can't show log window.";
static char LOG_FILE_OUT_OF_MEMORY[] = "Can't show log. Out of memory or too long log.";
#ifdef LOG_SHOW_ALWAYS
static char LOG_LIST_EMPTY[] = "      < empty >      ";
#endif

#define FATAL_ERROR_STRING "Fatal_error %s  %s\n"
#define ERROR_STRING       "Error       %s  %s\n"
#define WARNING_STRING     "Warning     %s  %s\n"
#define INFO_STRING        "Info        %s  %s\n"
#define EXINFO_STRING      "Ex_Info     %s  %s\n"

#define bufsize 1024 // max length for formated string



/* all names of modules should have same length (due to formating) */
static char modNames[][8] =
{
	"Errors ", "Main   ", "GUI    ", "Timers ",
	"Fonts  ", "Cockpit", "Input  ", "3dsLoad",
	"Scene  ", "SceneCl", "Texture", "TgaPcxL",
	"ArtifI ",
	"Unknown"
};



bool logFatalErrorsFound = false;
static bool logErrorsFound = false;
static bool logWarningsFound = false;
static FILE* F;


static void ShowErrorDlg(char *S);
static void ShowLog(FILE *O);




int logInit(void)
{
#ifdef USEMOD_LOG
	F = fopen(LOG_FILE_NAME, "wt");
	if(!F)
	{
		ShowErrorDlg(LOG_FILE_CREATE_ERROR);
		return 0;
	}

	logInfo(modErrors, "Log successfully started.");
#endif
	return 1;
}


int logDone(void)
{
#ifdef USEMOD_LOG
	logInfo(modErrors, "Log is shooting down.");
	if (fclose(F) != 0)
	{
		ShowErrorDlg(LOG_FILE_CLOSE_ERROR);
		return 0;
	}
#endif
	return 1;
}


void logShow(void)
{
#ifdef USEMOD_LOG
	FILE *O = fopen(LOG_FILE_NAME, "rt");
	if (O)
	{
		try
		{
			ShowLog(O);
		}
		catch (...) { fclose(O); throw; }
	}
	else ShowErrorDlg(LOG_FILE_READ_ERROR);
#endif
}


void logFatalError(logModule M, char *S, ...)
{
#ifdef USEMOD_LOG
	char buf[bufsize];
	va_list plist;
	assert(M <= modUnknown);

	logFatalErrorsFound = true;
	va_start(plist, S);
	vsnprintf(&buf[0], bufsize, S, plist);
	buf[bufsize-1] = '\0';
	va_end(plist);

	fprintf(F, FATAL_ERROR_STRING, modNames[M], buf);
#if defined(LOG_FLUSH_ALWAYS) || defined(LOG_FLUSH_WARNINGS) || defined(LOG_FLUSH_ERRORS)
	fflush(F);
#endif
#endif
}

//  ostream &logfile(enum logType e, enum module)
//  {
//    switch (e) {
//      case Info : if (infodisabled) return nullfile;
//      case Warning : cout << "WARNING "; break;
//      default :
//    }

//    return cout;
//  }

#ifndef DISABLE_LOG_ERRORS
void logError(logModule M, char *S, ...)
{
#ifdef USEMOD_LOG
  if (logLevelTable[M] >= logLevelErrors)
  {
	char buf[bufsize];
	va_list plist;
	assert(M <= modUnknown);

	logErrorsFound = true;
	va_start(plist, S);
	vsnprintf(&buf[0], bufsize, S, plist);
	buf[bufsize-1] = '\0';
	va_end(plist);

	//    log(Warning, modLogs) << "muj vektor: " << vec << endl;
	fprintf(F, ERROR_STRING, modNames[M], buf);
	#if defined(LOG_FLUSH_ALWAYS) || defined(LOG_FLUSH_WARNINGS) || defined(LOG_FLUSH_ERRORS)
		fflush(F);
	#endif
  }
#endif
}
#endif


#ifdef LOG_WARNINGS_ENABLED
void logWarning(logModule M, char *S, ...)
{
#ifdef USEMOD_LOG
  if (logLevelTable[M] >= logLevelWarnings)
  {
	char buf[bufsize];
	va_list plist;
	assert(M <= modUnknown);

	logWarningsFound = true;
	va_start(plist, S);
	vsnprintf(&buf[0], bufsize, S, plist);
	buf[bufsize-1] = '\0';
	va_end(plist);

	fprintf(F, WARNING_STRING, modNames[M], buf);
	#if defined(LOG_FLUSH_ALWAYS) || defined(LOG_FLUSH_WARNINGS)
	fflush(F);
	#endif
  }
#endif
}
#endif


#ifdef LOG_INFO_ENABLED
void logInfo(logModule M, char *S, ...)
{
#ifdef USEMOD_LOG
  if (logLevelTable[M] >= logLevelInfos)
  {
	char buf[bufsize];
	va_list plist;
	assert(M <= modUnknown);

	va_start(plist, S);
	vsnprintf(&buf[0], bufsize, S, plist);
	buf[bufsize-1] = '\0';
	va_end(plist);

	fprintf(F, INFO_STRING, modNames[M], buf);
	#if defined(LOG_FLUSH_ALWAYS)
		fflush(F);
	#endif
  }
#endif
}
#endif


#ifdef LOG_EXINFO_ENABLED
void logExInfo(logModule M, char *S, ...)
{
#ifdef USEMOD_LOG
  if (logLevelTable[M] >= logLevelExInfos)
  {
	char buf[bufsize];
	va_list plist;
	assert(M <= modUnknown);

	va_start(plist, S);
	vsnprintf(&buf[0], bufsize, S, plist);
	buf[bufsize-1] = '\0';
	va_end(plist);

	fprintf(F, EXINFO_STRING, modNames[M], buf);
	#if defined(LOG_FLUSH_ALWAYS)
		fflush(F);
	#endif
  }
#endif
}
#endif




static void ShowErrorDlg(char *S)
{
#ifdef WINDOZE
	MessageBox(NULL, S, "Error", MB_SYSTEMMODAL);
#endif
}


static void ShowLog(FILE *O)
{
#ifndef __linux__
								// Head !!!!!!!!!!!!!!!!!!!!
#ifndef LOG_SHOW_ALWAYS
	if (logFatalErrorsFound
	#ifdef LOG_SHOW_ON_WARNINGS
		|| logErrorsFound || logWarningsFound
	#else
		#ifdef LOG_SHOW_ON_ERROR
		|| logErrorsFound
		#endif
	#endif
		)
#endif
	{
		if (fseek(O, 0, SEEK_END) == 0)
		{
			fpos_t size;
			if (fgetpos(O, &size) == 0)
			{
				if (size == 0)
				{
				#ifdef LOG_SHOW_ALWAYS
				#ifdef WINDOZE
					MessageBox(NULL, &LOG_LIST_EMPTY[0], "Log list", MB_SYSTEMMODAL);
				#endif
				#endif
				}
				else
				{
					rewind(O);
					#ifndef BORLANDLIB // this is always true with Borland libraries
					if (size+1 <= UINT_MAX)
					#endif
					{
						size_t readed;
						char *S = (char*)malloc((unsigned int)(size+1));
						if (S != NULL)
						{
							try
							{
								if ((readed=fread(S, 1, (unsigned int)size, O)) >= 1)
								{
									S[readed] = '\0';
								#ifdef WINDOZE
									MessageBox(NULL, S, "Log list", MB_SYSTEMMODAL);
								#endif
								}
								else ShowErrorDlg(LOG_FILE_READ_ERROR);
							}
							catch (...) { free(S); throw; }
						}
						else ShowErrorDlg(LOG_FILE_OUT_OF_MEMORY);
					}
					#ifndef BORLANDLIB
					else ShowErrorDlg(LOG_FILE_OUT_OF_MEMORY);
					#endif
				}
			}
			else ShowErrorDlg(LOG_FILE_READ_ERROR);
		}
		else ShowErrorDlg(LOG_FILE_READ_ERROR);
	}
#endif
}



