/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */


#include "config.h"
#ifdef WINDOZE
#include <windows.h>
#endif
#include <string.h>
#pragma hdrstop // for C++ Builder compiler speed up

#include "Main.h"
#include "Log.h"
#include "Timers.h"
#include "gui.h"
#include "Cockpit.h"
#include "Fonts.h"
#include "3dsLoad.h"
#include "Scene.h"
#include "Input.h"
#include "Textures.h"
#include "RocketSmoke.h"
#include "Background.h"

#define WANT_CHDIR
#define WANT_STRDUP
#include "porting.h"

#include "setup.h"



int InitApp(int argc, char *argv[]);
void RunApp();
void DoneApp();
void SetAppDirectory(char **argv);



#ifndef WINDOZE
int main(int argc, char *argv[])
#else
WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
#endif
{
#ifdef USEMOD_MAIN
	try {
		// We need COMMAND-LINE params for windows platform
# ifdef WINDOZE
#  ifdef VISUAL_CPP
		// I can't find Visual Studio's functions to access command-line params
		int argc = 1;
		char **argv = &lpCmdLine;
#  else
		// Borland has _argc and _argv
		int argc = _argc;
		char **argv = _argv;
#  endif
# endif // WINDOZE

		// set CURRENT DIRECTORY
#if !defined(VISUAL_CPP) && !defined(MINGW) && !defined(UNIX)
		SetAppDirectory(argv);
#endif

		// MAIN code
		if (InitApp(argc, argv))
		{
			RunApp();
		}

	}

	// EXCEPTIONS catching
	catch (...)
	{
		logFatalError(modMain, "Application throwed unknown exception and will be terminated.");
	}
#endif // USEMOD_MAIN
	return 0;
}


#ifdef USEMOD_MAIN
static int DoneNeeded = 0;
#endif


int InitApp(int argc, char *argv[])
{
#ifdef USEMOD_MAIN
	if (!logInit())  { return 0; }

	DoneNeeded = 1;

	if (atexit(DoneApp) != 0)  { logFatalError(modMain, "Can't register atexit function"); DoneApp(); return 0; }


	if (!timInit())  { return 0; }
	if (!frInit())  { return 0; }
	if (!guiInit(argc, argv))  { return 0; }
	if (!FontsInit())  { return 0; }
	if (!CockpitInit())  { return 0; }
	if (!Load3DSInit())  { return 0; }
	if (!InputInit())  { return 0; }
	if (!TexturesInit())  { return 0; }
	if (!SceneInit())  { return 0; }
	if (!BackgroundInit())  { return 0; }
	if (!RocketSmokeInit())  { return 0; }

	logInfo(modMain, "All modules started successfully.");
#endif
	return 1;
}


void RunApp()
{
#ifdef USEMOD_MAIN
	guiMainLoop();
#endif
}


void DoneApp()
{
#ifdef USEMOD_MAIN

#define SafeShootDown(A,N) try {A;} catch(...) { logError(modMain, "Module %s throwed exception during shooting down.", N); }
	try {
		if (DoneNeeded) {
			DoneNeeded = 0;
			logInfo(modMain, "Starting modules shooting down.");

			SafeShootDown(RocketSmokeDone(),"RocketSmoke");
			SafeShootDown(BackgroundDone(),"Background");
			SafeShootDown(SceneDone(),    "Scene");
			SafeShootDown(TexturesDone(), "Textures");
			SafeShootDown(InputDone(),    "Input");
			SafeShootDown(Load3DSDone(),  "3dsLoad");
			SafeShootDown(CockpitDone(),  "Cockpit");
			SafeShootDown(FontsDone(),    "Fonts");
			SafeShootDown(guiDone(),      "GUI");

			/* TODO : CheckMemoryLeaks */

			logInfo(modMain, "All modules shooted successfully.");
			try {
		if (logDone()) logShow();
			} catch (...) {};
		}
	} catch(...) {};
#endif
}



void SetAppDirectory(char **argv)
{
#ifdef USEMOD_MAIN
	char *path = strdup(argv[0]);
	if (!path) {
		fprintf(stderr, "Not enough memory.\n");
		exit(0);
	}
# ifdef WINDOZE
	char *c = strrchr(path , '\\');
# else
	char *c = strrchr(path , '/');
# endif
	if (*c) {
		*c = 0;
		if (chdir(path) != 0)
		{
			fprintf(stderr, "Can't find directory \"%s\".\n", path);
			free(path);
			exit(0);
		}
	} else {
		fprintf(stderr, "Bad path \"%s\".\n", argv[0]);
		exit(0);
	}
#endif
}



