/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "TgaPcxLoad.h"
#include "Log.h"
#include "porting.h"




#define BYTE unsigned char
#define WORD unsigned int
#define DWORD unsigned long
#define CHAR char
#define SHORT short int
#define BOOL int
#define VOID void




/* TGA File Header */
typedef struct _TgaHeader
{
    BYTE   IdLength;            /* Image ID Field Length      */
    BYTE   CmapType;            /* Color Map Type             */
    BYTE   ImageType;           /* Image Type                 */

    /* Color Map Specification   */
    WORD   CmapIndex;           /* First Entry Index          */
    WORD   CmapLength;          /* Color Map Length           */
    BYTE   CmapEntrySize;       /* Color Map Entry Size       */

    /* Image Specification */
    WORD   X_Origin;            /* X-origin of Image          */
    WORD   Y_Origin;            /* Y-origin of Image          */
    WORD   ImageWidth;          /* Image Width                */
    WORD   ImageHeight;         /* Image Height               */
    BYTE   PixelDepth;          /* Pixel Depth                */
    BYTE   ImagDesc;            /* Image Descriptor           */
} TGAHEADER;


/* Image/Color Map Data */
typedef struct _TgaColorMap
{
    CHAR  *IdField;             /* Image ID Field             */
    BYTE  *CmapData;            /* Color Map Data             */
} TGACMAP;


/* Developer Area Tag Structure */

typedef struct _TgaTag
{
    WORD    TagNumber;           /* Number of the Tag          */
    DWORD   TagOffset;           /* Offset of the Tag Data     */
    DWORD   TagSize;             /* Size of the Tag Data       */
    BYTE   *TagData;             /* Pointer to the Tag Data    */
    struct _TgaTag *Next;        /* Link to next Tag           */
} TGATAG;


/* Developer Area */
typedef struct _TgaDeveloper
{
    WORD    NumberOfTags;     /* Number of Tags in Directory  */
    TGATAG *TagList;          /* Link to list of Tags         */
} TGADEVELOPER;


/* Extension Area */
typedef struct _TgaExtension
{
    WORD   Size;                /* Extension Size             */
    CHAR   AuthorName[41];      /* Author Name                */
    CHAR   AuthorComment[324];  /* Author Comment             */
    WORD   StampMonth;          /* Date/Time Stamp: Month     */
    WORD   StampDay;            /* Date/Time Stamp: Day       */
    WORD   StampYear;           /* Date/Time Stamp: Year      */
    WORD   StampHour;           /* Date/Time Stamp: Hour      */
    WORD   StampMinute;         /* Date/Time Stamp: Minute    */
    WORD   StampSecond;         /* Date/Time Stamp: Second    */
    CHAR   JobName[41];         /* Job Name/ID                */
    WORD   JobHour;             /* Job Time: Hours            */
    WORD   JobMinute;           /* Job Time: Minutes          */
    WORD   JobSecond;           /* Job Time: Seconds          */
    CHAR   SoftwareId[41];      /* Software ID                */
    WORD   VersionNumber;       /* Software Version Number    */
    BYTE   VersionLetter;       /* Software Version Letter    */
    DWORD  KeyColor;            /* Key Color                  */
    WORD   PixelNumerator;      /* Pixel Aspect Ratio Numerator   */
    WORD   PixelDenominator;    /* Pixel Aspect Ratio Denominator */
    WORD   GammaNumerator;      /* Gamma Value                */
    WORD   GammaDenominator;    /* Gamma Value                */
    DWORD  ColorOffset;         /* Color Correction Offset    */
    DWORD  StampOffset;         /* Postage Stamp Offset       */
    DWORD  ScanOffset;          /* Scan Line Table Offset     */
    BYTE   AttributesType;      /* Attributes Types           */
    DWORD *ScanLineTable;       /* Scan Line Table            */
    BYTE   StampWidth;          /* Width of postage stamp image */
    BYTE   StampHeight;         /* Height of postage stamp image */
    BYTE  *StampImage;          /* Postage Stamp Image        */
    WORD   ColorTable[1024];    /* Color Correction Table     */
} TGAEXTENSION;


/* TGA File Footer */
typedef struct _TgaFooter
{
    DWORD ExtensionOffset;      /* Extension Area Offset      */
    DWORD DeveloperOffset;      /* Developer Directory Offset */
    CHAR  Signature[18];        /* Signature, dot, and NULL   */
} TGAFOOTER;


/* Complete TGA File Format */
typedef struct _TgaFormat
{
    TGAHEADER      Head;            /* Header Area                          */
    TGACMAP        Cmap;            /* Image/Color Map Area                 */
    TGADEVELOPER   Developer;       /* Developer Area                       */
    TGAEXTENSION   Extension;       /* Extension Area                       */
    TGAFOOTER      Foot;            /* File Footer Area                     */
    BOOL           NewTgaFormat;    /* TRUE if file is v2.0 TGA format      */
    DWORD          TgaDataOffset;   /* Offset of the image data in the file */
} TGA;

#define TGASIGNATURE   "TRUEVISION-XFILE.\0"







/*
**  The PCX header format.
*/
typedef struct _PcxHeader       /* Offset   Description            */
{
    BYTE   Id;                  /*  00h     Manufacturer ID        */
    BYTE   Version;             /*  01h     Version                */
    BYTE   Format;              /*  02h     Encoding Scheme        */
    BYTE   BitsPixelPlane;      /*  03h     Bits/Pixel/Plane       */
    WORD   Xmin;                /*  04h     X Start (upper left)   */
    WORD   Ymin;                /*  06h     Y Start (top)          */
    WORD   Xmax;                /*  08h     X End (lower right)    */
    WORD   Ymax;                /*  0Ah     Y End (bottom)         */
    WORD   Hdpi;                /*  0Ch     Horizontal Resolution  */
    WORD   Vdpi;                /*  0Eh     Vertical Resolution    */
    BYTE   EgaPalette[48];      /*  10h     16-Color EGA Palette   */
    BYTE   Reserved;            /*  40h     Reserved               */
    BYTE   NumberOfPlanes;      /*  41h     Number of Color Planes */
    WORD   BytesLinePlane;      /*  42h     Bytes/Line/Plane       */
    WORD   PaletteInfo;         /*  44h     Palette Interpretation */
    WORD   HScreenSize;         /*  46h     Horizontal Screen Size */
    WORD   VScreenSize;         /*  48h     Vertical Screen Size   */
    BYTE   Filler[54];          /*  4Ah     Reserved               */
} PCXHEADER;

/*
**  PCX VGA palette.
*/
/*typedef struct _PcxVgaPalette
{
    BYTE   VgaPalette[768];     /*          256 VGA Color Palette  */
//} PCXVGAPALETTE;

/*
**  The DCX header format.
*/
//typedef struct _DcxHeader       /* Offset   Description            */
//{
//    DWORD   Id;                 /*   0h     File ID                */
//    DWORD   PageList[1024];     /*   4h     Page List Array        */
//} DCXHEADER;








int TgaInit(TEXINFO &T)
{
	T.HeaderLoaded = false;
	T.BodyLoaded = false;

	T.IntData = malloc(sizeof(TGA));
	if (T.IntData == NULL)
	{
		logError(modTgaPcxLoad, "Out of memory.");
		return 0;
	}

	return 1;
}


void TgaDone(TEXINFO &T)
{
	free(T.IntData);
	T.IntData = NULL;
}







/* Reads word from file */
WORD GetWord(FILE *FpTga)
{
  /* TODO : Endians */
  return((getc(FpTga))+(getc(FpTga)<<8));
}

/* Reads dword from file */
DWORD GetDword(FILE *FpTga)
{
  /* TODO : Endians */
  return((getc(FpTga))+(getc(FpTga)<<8)+(getc(FpTga)<<16)+(getc(FpTga)<<24));
}


/************************************************************************/
/************************************************************************/
/*									*/
/*				   TGA					*/
/*									*/
/************************************************************************/
/************************************************************************/

void FreeTagList(TGATAG *Head)
{
	TGATAG *xtag1, *xtag2;
	for (xtag1 = Head; xtag1!=NULL; )
	{
		xtag2 = xtag1->Next;
		free(xtag1);
		xtag1 = xtag2;
	}
}

void FreeHeader(TGA *TgaHead)
{
	free(TgaHead->Cmap.IdField);
	free(TgaHead->Cmap.CmapData);
	FreeTagList(TgaHead->Developer.TagList);
	free(TgaHead->Extension.ScanLineTable);
	free(TgaHead->Extension.StampImage);
}

/* Read the information from a TGA file info a TGAHEADER structure. */
int ReadTgaHeader(TGA *TgaHead,FILE *FpTga)
     /*TGA  *TgaHead;              - Pointer to TGA header structure              */
     /*FILE *FpTga;                - TGA image file input FILE stream             */
/* returns: 1-success, 0-out of memory */
{
  WORD    i, j;           /* Loop counters                                */
  WORD    cmapsize;       /* Size of the color map in bytes               */
  WORD    stampsize;      /* Size of the postage stamp image in bytes     */
  TGATAG *tag;            /* TGA tag structure                            */
  TGATAG *head;           /* Head of TGA tag linked list                  */
  TGATAG *current_tag;    /* Current node of linked list                  */

	/* Read the TGA header (Version 1.0 and 2.0). */
	TgaHead->Head.IdLength      = getc(FpTga);
	TgaHead->Head.CmapType      = getc(FpTga);
	TgaHead->Head.ImageType     = getc(FpTga);
	TgaHead->Head.CmapIndex     = GetWord(FpTga);
	TgaHead->Head.CmapLength    = GetWord(FpTga);
	TgaHead->Head.CmapEntrySize = getc(FpTga);
	TgaHead->Head.X_Origin      = GetWord(FpTga);
	TgaHead->Head.Y_Origin      = GetWord(FpTga);
	TgaHead->Head.ImageWidth    = GetWord(FpTga);
	TgaHead->Head.ImageHeight   = GetWord(FpTga);
	TgaHead->Head.PixelDepth    = getc(FpTga);
	TgaHead->Head.ImagDesc      = getc(FpTga);

	/* Read the Image ID field. */
	if (TgaHead->Head.IdLength)
	{
		/* Allocate memory for the Id Field data */
		if ((TgaHead->Cmap.IdField =
			(CHAR *) calloc(TgaHead->Head.IdLength, sizeof(CHAR))) ==
			(CHAR *) NULL)
		{
			logError(modTgaPcxLoad, "Out of memory.");
			return 0;
		}

		/* Read the Image ID data */
		/* Do not assume that it is a NULL-terminated string */
		for (i = 0; (signed)i < TgaHead->Head.IdLength; i++)
			TgaHead->Cmap.IdField[i] = (CHAR) getc(FpTga);
	}
	else
		TgaHead->Cmap.IdField = (CHAR *) NULL;

	/* Read the color map data (Version 1.0 and 2.0). */
	if (TgaHead->Head.CmapType)
	{
		/* Determine the size of the color map */
		cmapsize = ((TgaHead->Head.CmapEntrySize + 7) >> 3) *
		TgaHead->Head.CmapLength;

		/* Allocate memory for the color map data */
		if ((TgaHead->Cmap.CmapData =
			(BYTE *) calloc(cmapsize, sizeof(BYTE))) == (BYTE *) NULL)
		{
			/* clean up */
			free(TgaHead->Cmap.IdField);
			logError(modTgaPcxLoad, "Out of memory.");
			return 0;
		}

		/* Read the color map data */
		for (i = 0; i < cmapsize; i++)
			TgaHead->Cmap.CmapData[i] = getc(FpTga);
	}
	else
		TgaHead->Cmap.CmapData = (BYTE *) NULL;

	// initalize some pointers to NULL
	TgaHead->Developer.TagList = NULL;
	TgaHead->Extension.ScanLineTable = NULL;
	TgaHead->Extension.StampImage = NULL;


  /*
  ** Store the offset of the image data.  This field is not part of
  ** the TGA format, but it helps the software locate the data when
  ** reading/writing TGA files.
  */
  TgaHead->TgaDataOffset = ftell(FpTga);

  /* Check the version of the TGA file (Versions 1.0 and 2.0).  */
  fseek(FpTga, -26, SEEK_END);  /* Seek to the (possible) TGA footer */

  /* Read in the (possible) offset values */
  TgaHead->Foot.ExtensionOffset = GetDword(FpTga);
  TgaHead->Foot.DeveloperOffset = GetDword(FpTga);

  /* Read in the (possible) signature */
  for (i = 0; i < sizeof(TgaHead->Foot.Signature); i++)
    TgaHead->Foot.Signature[i] = (CHAR) getc(FpTga);

  /* Check if the data read is a TGA signature string */
  if (!strcmp(TgaHead->Foot.Signature, TGASIGNATURE))
    TgaHead->NewTgaFormat = 1;   /* Yes.  Version 2.0 */
  else
    TgaHead->NewTgaFormat = 0;   /* No.  Version 1.0 */

  /* If version is 2.0 then check for developer and extension areas. */
  if (TgaHead->NewTgaFormat == 1)
  {
    /* Check for the presence of a developers area */
    if (TgaHead->Foot.DeveloperOffset)
    {
      /* Seek to the developer area */
      fseek(FpTga, TgaHead->Foot.DeveloperOffset, SEEK_SET);

      /* Get the number of tags in the directory */
      TgaHead->Developer.NumberOfTags = GetWord(FpTga);

      /* Read the tags and store as a singly-linked list */
      head = (TGATAG *) NULL;
      for (i = 0; i < TgaHead->Developer.NumberOfTags; i++)
      {
        /* Allocate the tag */
        if ((tag = (TGATAG *) malloc(sizeof(TGATAG))) == (TGATAG *) NULL)
        {
				/* clean up */
				free(TgaHead->Cmap.IdField);
				free(TgaHead->Cmap.CmapData);
				FreeTagList(head);
				logError(modTgaPcxLoad, "Out of memory.");
					return 0;
        }

        /* Read the tag information */
        tag->TagNumber = GetWord(FpTga);
        tag->TagOffset = GetDword(FpTga);
        tag->TagSize   = GetDword(FpTga);

        /* Read the tag data */
        /* Seek to the tag data */
        fseek(FpTga, tag->TagOffset, SEEK_SET);

        /* Read the tag data */
        for (j = 0; j < tag->TagSize; j++)
            tag->TagData[j] = getc(FpTga);

        /* Link the tag to the list */
        if (head == (TGATAG *) NULL)
        {
          head = tag;
          current_tag = head;
        }
        else
        {
          current_tag->Next = tag;
          tag->Next = (TGATAG *) NULL;
          current_tag = tag;
        }
      }
      TgaHead->Developer.TagList = head;  /* Assign list to structure */
    }

    /* Check for the presence of an extension area */
    if (TgaHead->Foot.ExtensionOffset)
    {
      /* Seek to the extension area */
      fseek(FpTga, TgaHead->Foot.ExtensionOffset, SEEK_SET);

      /* Read the extension area size information */

      TgaHead->Extension.Size = GetWord(FpTga);

      /* Read the extension information defined in TGA version 2.0 */
      for (i = 0; i < sizeof(TgaHead->Extension.AuthorName); i++)
        TgaHead->Extension.AuthorName[i] = (CHAR) getc(FpTga);

      for (i = 0; i < sizeof(TgaHead->Extension.AuthorComment); i++)
        TgaHead->Extension.AuthorComment[i] = (CHAR) getc(FpTga);

      TgaHead->Extension.StampMonth  = GetWord(FpTga);
      TgaHead->Extension.StampDay    = GetWord(FpTga);
      TgaHead->Extension.StampYear   = GetWord(FpTga);
      TgaHead->Extension.StampHour   = GetWord(FpTga);
      TgaHead->Extension.StampMinute = GetWord(FpTga);
      TgaHead->Extension.StampSecond = GetWord(FpTga);

      for (i = 0; i < sizeof(TgaHead->Extension.JobName); i++)
        TgaHead->Extension.JobName[i] = (CHAR) getc(FpTga);

      TgaHead->Extension.JobHour     = GetWord(FpTga);
      TgaHead->Extension.JobMinute   = GetWord(FpTga);
      TgaHead->Extension.JobSecond   = GetWord(FpTga);

      for (i = 0; i < sizeof(TgaHead->Extension.SoftwareId); i++)
        TgaHead->Extension.SoftwareId[i] = (CHAR) getc(FpTga);

      TgaHead->Extension.VersionNumber    = GetWord(FpTga);
      TgaHead->Extension.VersionLetter    = getc(FpTga);
      TgaHead->Extension.KeyColor         = GetDword(FpTga);
      TgaHead->Extension.PixelNumerator   = GetWord(FpTga);
      TgaHead->Extension.PixelDenominator = GetWord(FpTga);
      TgaHead->Extension.GammaNumerator   = GetWord(FpTga);
      TgaHead->Extension.GammaDenominator = GetWord(FpTga);
      TgaHead->Extension.ColorOffset      = GetDword(FpTga);
      TgaHead->Extension.StampOffset      = GetDword(FpTga);
      TgaHead->Extension.ScanOffset       = GetDword(FpTga);
      TgaHead->Extension.AttributesType   = getc(FpTga);

      /* Check for the presence of a scan line table */
      if (TgaHead->Extension.ScanOffset)
      {
        /* Seek to the postage scan line table */
        fseek(FpTga, TgaHead->Extension.ScanOffset, SEEK_SET);

        /* Allocate memory.  One DWORD per line in image */
        if ((TgaHead->Extension.ScanLineTable =
            (DWORD *) calloc(TgaHead->Head.ImageHeight, sizeof(DWORD))) ==
            (DWORD *) NULL)
        {
          /* clean up */
          free(TgaHead->Cmap.IdField);
          free(TgaHead->Cmap.CmapData);
          FreeTagList(TgaHead->Developer.TagList);

		      logError(modTgaPcxLoad, "Out of memory.");
					return 0;
        }

        /* Read in scan line offset values */
        for (i = 0; i < TgaHead->Head.ImageHeight; i++)
          TgaHead->Extension.ScanLineTable[i] = GetDword(FpTga);
      }

      /* Check for the presence of a postage stamp image*/
      if (TgaHead->Extension.StampOffset)
      {
        /* Seek to the postage stamp image */
        fseek(FpTga, TgaHead->Extension.StampOffset, SEEK_SET);

        /* Read the size of the stamp in pixels */
        TgaHead->Extension.StampWidth  = getc(FpTga);
        TgaHead->Extension.StampHeight = getc(FpTga);

        /* Calculate the size of the stamp in bytes */
        stampsize = TgaHead->Extension.StampWidth *
                    TgaHead->Extension.StampHeight *
                    ((TgaHead->Head.PixelDepth + 7) >> 3);

        /* Allocate memory for the postage stamp image */
        if ((TgaHead->Extension.StampImage =
            (BYTE *) calloc(stampsize, sizeof(BYTE))) ==
            (BYTE *) NULL)
        {
				/* clean up */
				free(TgaHead->Cmap.IdField);
				free(TgaHead->Cmap.CmapData);
				FreeTagList(TgaHead->Developer.TagList);
				free(TgaHead->Extension.ScanLineTable);

				logError(modTgaPcxLoad, "Out of memory.");
					return 0;
        }

        /* Read the stamp data one byte at a time */
        for (i = 0; i < stampsize; i++)
          TgaHead->Extension.StampImage[i] = getc(FpTga);
      }

      /* Check for the presence of a color correction table */
      if (TgaHead->Extension.ColorOffset)
      {
        /* Seek to the color correction table */
        fseek(FpTga, TgaHead->Extension.ColorOffset, SEEK_SET);

        /* Read in the entire 1024 WORD (2048 byte) table */
        for (i = 0; i < sizeof(TgaHead->Extension.ColorTable); i++)
          TgaHead->Extension.ColorTable[i] = GetWord(FpTga);
      }
    }
  }

  return 1;
}



/*  Read encoded data from a TGA image file.*/
int TgaDecodeScanLine(BYTE *DecodedBuffer,WORD LineLength,WORD PixelSize,FILE *FpTga)
     /* BYTE *DecodedBuffer;      - Pointer to buffer to hold decoded data         */
     /* WORD  LineLength;         - The length of a scan line in pixels            */
     /* WORD  PixelSize;          - The number of bytes in a pixel                 */
     /* FILE *FpTga;              - FILE pointer to the open input TGA image file  */
/* returns: -1 - Pixel run will overflow buffer
            byteCount - Always positive (>=0), success*/
{
  WORD    i;              /* Loop counter                                 */
  SHORT   byteCount;      /* Number of bytes written to the buffer        */
  WORD    runCount;       /* The pixel run count                          */
  WORD    bufIndex;       /* The index of DecodedBuffer                   */
  WORD    bufMark;        /* Index marker of DecodedBuffer                */
  WORD    pixelCount;     /* The number of pixels read from the scan line */

  bufIndex   = 0;         /* Initialize buffer index  */
  byteCount  = 0;         /* Initialize byte count    */
  pixelCount = 0;         /* Initialize pixel counter */

  /* Main decoding loop */
  while (pixelCount < LineLength)
  {
    /* Get the pixel count */
    runCount = getc(FpTga);

    /* Make sure writing this next run will not overflow the buffer */
    if (pixelCount + (runCount & 0x7f) + 1 > LineLength)
		{ logError(modTgaPcxLoad, "Pixel run overflow. File is corrupted."); return(-1); }  /* Pixel run will overflow buffer */

	/* If the run is encoded... */
    if (runCount & 0x80)
    {
      runCount &= ~0x80;              /* Mask off the upper bit       */

      bufMark = bufIndex;             /* Save the start-of-run index  */

      /* Update total pixel count */
      pixelCount += (runCount + 1);

      /* Update the buffer byte count */
      byteCount += ((runCount + 1) * PixelSize);

      /* Write the first pixel of the run to the buffer */
      bufIndex += PixelSize;
      for (i = 1; i <= PixelSize; i++)
        DecodedBuffer[bufIndex-i] = getc(FpTga);
//        DecodedBuffer[bufIndex++] = getc(FpTga);

      /* Write remainder of pixel run to buffer 'runCount' times */
      while (runCount--)
      {
        for (i = 0; i < PixelSize; i++)
          DecodedBuffer[bufIndex++] = DecodedBuffer[bufMark + i];
      }
    }
    else    /* ...the run is unencoded (raw) */
    {
      /* Update total pixel count */
      pixelCount += (runCount + 1);

      /* Update the buffer byte count */
      byteCount  += ((runCount + 1) * PixelSize);

      /* Write runCount pixels */
      do
      {
        bufIndex += PixelSize;
        for (i = 1; i <= PixelSize; i++)
          DecodedBuffer[bufIndex-i] = getc(FpTga);
//          DecodedBuffer[bufIndex++] = getc(FpTga);
      }
      while (runCount--);
    }
  }
  return(byteCount);
}


/************************************************************************/
/************************************************************************/
/*									*/
/*				   PCX					*/
/*									*/
/************************************************************************/
/************************************************************************/


VOID ReadPcxHeader(PCXHEADER *PcxHead,FILE *FpPcx)
     /* PCXHEADER *PcxHead;     - Pointer to PCX header structure  */
     /* FILE      *FpPcx;       - PCX image file input FILE stream */
{
    register SHORT i;

    PcxHead->Id             = getc(FpPcx);
    PcxHead->Version        = getc(FpPcx);
    PcxHead->Format         = getc(FpPcx);
    PcxHead->BitsPixelPlane = getc(FpPcx);
    PcxHead->Xmin           = GetWord(FpPcx);
    PcxHead->Ymin           = GetWord(FpPcx);
    PcxHead->Xmax           = GetWord(FpPcx);
    PcxHead->Ymax           = GetWord(FpPcx);
    PcxHead->Hdpi           = GetWord(FpPcx);
    PcxHead->Vdpi           = GetWord(FpPcx);

    /* Read the EGA Palette */
    for (i = 0; i < sizeof(PcxHead->EgaPalette); i++)
        PcxHead->EgaPalette[i] = getc(FpPcx);

    PcxHead->Reserved       = getc(FpPcx);       
    PcxHead->NumberOfPlanes = getc(FpPcx);
    PcxHead->BytesLinePlane = GetWord(FpPcx);
    PcxHead->PaletteInfo    = GetWord(FpPcx);  
    PcxHead->HScreenSize    = GetWord(FpPcx);
    PcxHead->VScreenSize    = GetWord(FpPcx);

    /* Read the reserved area at the end of the header */
    for (i = 0; i < sizeof(PcxHead->Filler); i++)
        PcxHead->Filler[i] = getc(FpPcx);
}





SHORT PcxDecodeScanLine(BYTE *DecodedBuffer, WORD BufferSize, FILE *FpPcx)
     /* BYTE *DecodedBuffer;    - Pointer to buffer to hold decoded data         */
     /* WORD  BufferSize;       - Size of buffer to hold decoded data            */
     /* FILE *FpPcx;            - FILE pointer to the open input PCX image file  */
{
    WORD  index = 0;            /* Index into compressed scan line buffer */
    SHORT total = 0;            /* Running total of decoded pixel values  */
    BYTE  byte;                 /* Data byte read from PCX file           */
    static BYTE runcount = 0;   /* Length of decoded pixel run            */
    static BYTE runvalue = 0;   /* Value of decoded pixel run             */

    /*
    ** If there is any data left over from the previous scan
    ** line write it to the beginning of this scan line.
    */
    do
    {
        /* Write the pixel run to the buffer */
        for (total += runcount;                 /* Update total             */
             runcount && index < BufferSize;    /* Don't read past buffer   */
             runcount--, index++)
            DecodedBuffer[index] = runvalue;    /* Assign value to buffer   */

        if (runcount)           /* Encoded run ran past end of scan line    */
        {
            total -= runcount;  /* Subtract count not written to buffer     */
            return(total);      /* Return number of pixels decoded          */
        }

        /*
        ** Get the next encoded run packet.
        **
        ** Read a byte of data.  If the two MBSs are 1 then this byte
        ** holds a count value (0 to 63) and the following byte is the
        ** data value to be repeated.  If the two MSBs are 0 then the
        ** count is one and the byte is the data value itself.
        */
        byte = getc(FpPcx);                  /* Get next byte    */

        if ((byte & 0xC0) == 0xC0)              /* Two-byte code    */
        {
            runcount = byte & 0x3F;             /* Get run count    */
            runvalue = getc(FpPcx);          /* Get pixel value  */
        }
        else                                    /* One byte code    */
        {
            runcount = 1;                       /* Run count is one */
            runvalue = byte;                    /* Pixel value      */
        }
    }
    while (index < BufferSize);     /* Read until the end of the buffer     */

    if (ferror(FpPcx))
        return(EOF);    /* File stream error.  Probably a premature EOF     */

    return(total);      /* Return number of pixels decoded                  */
}

/************************************************************************/
/************************************************************************/
/*									*/
/*				MAIN FUNCTIONS				*/
/*									*/
/************************************************************************/
/************************************************************************/

int TgaReadHeader(TEXINFO &T, char *FileName)
{
  TGA  *tgaHead = (TGA*)T.IntData;

  assert(T.HeaderLoaded == false);

  char *name = strdup(FileName);
  convertFilename(name);
  T.F       = fopen(name, "rb");
  free(name);

  if (T.F==NULL)
  {
    logError(modTgaPcxLoad, "Can't open file \"%s\".", FileName);
    return 0;
  }

  /* Read the TGA image file header information */
  if (!ReadTgaHeader(tgaHead, T.F))
  {
    logError(modTgaPcxLoad, "Can't read file header (\"%s\").", FileName);
    fclose(T.F);
    return 0;
  }

  /* Check for FILE stream error */
  if (ferror(T.F))
  {
    logError(modTgaPcxLoad, "File stream error while reading header information.");
    fclose(T.F);
    FreeHeader(tgaHead);
    return 0;
  }

  /* Seek to the image data */
  fseek(T.F, tgaHead->TgaDataOffset, SEEK_SET);

  /* Image data is encoded */
  if (!(tgaHead->Head.ImageType == 9  || tgaHead->Head.ImageType == 10 ||
	      tgaHead->Head.ImageType == 11 || tgaHead->Head.ImageType == 1  ||
        tgaHead->Head.ImageType == 2  || tgaHead->Head.ImageType == 3))
  {
		logError(modTgaPcxLoad, "Unknown image type %d.", tgaHead->Head.ImageType);
    fclose(T.F);
    FreeHeader(tgaHead);
    return 0;
  }

  T.Width = tgaHead->Head.ImageWidth;
  T.Height = tgaHead->Head.ImageHeight;
  #define BYTESPERPIXEL   ((tgaHead->Head.PixelDepth + 7) >> 3)
  T.BPP = BYTESPERPIXEL;
  switch (T.BPP)
  {
  	case 3 : T.Format = GL_RGB; break;
    case 4 : T.Format = GL_RGBA; break;
    default: { logError(modTgaPcxLoad, "Unsupported PixelDepth: %d.", tgaHead->Head.PixelDepth); return 0; }
  }

  logExInfo(modTgaPcxLoad, "TGA image type: %d.", tgaHead->Head.ImageType);

  T.HeaderLoaded = true;

  return 1;
}


int TgaReadBody(TEXINFO &T, void *Buf)
{
  WORD        i;              /* Loop counters                            */
  DWORD       imageSize;      /* Size of the image data in bytes          */
  DWORD       bufSize;        /* Size of the scan line buffer             */
  SHORT       byteCount;      /* Number of bytes in a buffer              */
  DWORD	      bytesPerLine;
  TGA         *tgaHead = (TGA*)T.IntData;


	assert(T.HeaderLoaded == true);
  assert(T.BodyLoaded == false);

  bytesPerLine=(tgaHead->Head.ImageWidth*T.BPP);
  bufSize = (bytesPerLine * tgaHead->Head.ImageHeight);
//  *xRes=tgaHead->Head.ImageWidth;
//  *yRes=tgaHead->Head.ImageHeight;

  imageSize = 0;

  /* Decode a scan line and write it to the temporary file */
  if (tgaHead->Head.ImageType == 1 || tgaHead->Head.ImageType == 2 ||
      tgaHead->Head.ImageType == 3)
  {
    fread(Buf, bufSize, 1, T.F);
  }
  else
  {
    for (i = 0; i < tgaHead->Head.ImageHeight; i++)
    {
      /* Decode a scan line */
      if ((byteCount = TgaDecodeScanLine(&(((unsigned char*)Buf)[bytesPerLine*i]),
            tgaHead->Head.ImageWidth, BYTESPERPIXEL, T.F)) < 0)
      {
				fclose(T.F);
        FreeHeader(tgaHead);
        return 0;
      }
      /* Update the image data size byte count */
      imageSize += byteCount;
    }
  }
  /*printf("\n\nTGA file decoded to file \n" );*/

	fclose(T.F);
  FreeHeader(tgaHead);

  T.BodyLoaded = true;

  return 1;
}




int decodePcx(FILE *fpPcxIn,BYTE **decodedBuffer, WORD *xRes, WORD *yRes)
{
    register WORD i;                /* Loop counter                      */
    PCXHEADER     pcxHead;          /* PCX image file header structure   */
    WORD          numOfColors;      /* Number of colors in the image     */
    WORD          xSize;            /* Width of image in pixels          */
    WORD          ySize;            /* Height of image in scan lines     */
    SHORT         scanLineLength;       /* Length of uncompressed scan line  */
    /*    WORD          scanLineSize;     Size of encoded scan line         */
    SHORT         ret;              /* Return value holder               */

    puts("PCXCODE - Decode and recode a PCX image file (v1.00)\n");

    /* Read the PCX image file header information */
    ReadPcxHeader(&pcxHead, fpPcxIn);

    /* Check for FILE stream error */
    if (ferror(fpPcxIn))
    {
        fputs("PCXHEAD: Error reading header information!\n", stderr);
        exit(-4);
    }

    /* Check the identification byte value */
    if (pcxHead.Id != 0x0A)
    {
        fprintf(stderr, "PCXHEAD: Not a PCX-format file!\n");
        exit(-5);
    }

    /* Calculate size of image in pixels and scan lines */
    xSize = pcxHead.Xmax - pcxHead.Xmin + 1;
    ySize = pcxHead.Ymax - pcxHead.Ymin + 1;

    /* Do some more calculations */
    numOfColors = (1 << (pcxHead.BitsPixelPlane * pcxHead.NumberOfPlanes));
    scanLineLength = pcxHead.NumberOfPlanes * pcxHead.BytesLinePlane;

    /* Display some PCX image file information */
    printf("Image Width  = %3d\nImage Length = %3d\n", xSize, ySize);
    printf("Number of Colors: %d\n", numOfColors);
    printf("Size of uncompressed scan line: %d\n\n", scanLineLength);

    /* Create an uncompressed scanline buffer */
    if ((*decodedBuffer = (BYTE *) malloc(scanLineLength*ySize)) == NULL)
    {
        fputs("PCXCODE: Cannot allocate scan line buffer.", stderr);
        exit(-6);
    }

    *xRes=xSize;
    *yRes=ySize;

    for (i = 0; i < ySize; i++)
    {
        printf("Scan Line %03d\r", i);

        /* Read scan line and decode it */
        if ((ret = PcxDecodeScanLine(&((*decodedBuffer)[scanLineLength*i]),
             scanLineLength, fpPcxIn)) != scanLineLength)
        {
            printf("PCXCODE: Bad scan line length (was %d should be %d)\n",
              ret, scanLineLength);
        }

    }

    /* If there is a VGA palette then read and append it to the output file */
/*    if (pcxHead.Version == 5)
    {
        fseek(fpPcxIn, -769L, SEEK_END);  // Seek backwards from end of file

        if (GetByte(fpPcxIn) == 0x0C)     // VGA Palette ID value?
        {
            PCXVGAPALETTE  PcxVgaPal;     // PCX VGA palette structure

            fputc(0x0c, fpPcxOut);

            // Read the VGA palette
            for (i = 0; i < sizeof(PcxVgaPal); i++)
                PcxVgaPal.VgaPalette[i] = GetByte(fpPcxIn);

            // Write the VGA palette
            for (i = 0; i < sizeof(PcxVgaPal); i++)
                PutByte(PcxVgaPal.VgaPalette[i], fpPcxOut);
        }
    }         */

    return(0);
}


#ifdef xxxx
int makeTex(GLuint *TexID, const char *fileName)
/* returns: makeTexSuccess, makeTexOutOfMem, makeTexErrorInInputFile,
            makeTexFileNotFound, makeTexUnknownFileExt, makeTexGLUOutOfMem,
            makeTexGLUError */
{
  FILE  *fpIn;          /* image file input FILE stream  */
  BYTE  *buffer;
  WORD  xRes, yRes;
  char  *tmp;
  char  pripona[5];
  int   r;

  /* Open the input image file */
  convertFilename(fileName);
  if ((fpIn = fopen(fileName, "rb")) == NULL)
  {
    /*fprintf(stderr, "TextureLoader: Cannot open file %s\n", fileName);*/
    return makeTexFileNotFound;
  }

  /* Extension from fileName */
  tmp = strrchr(fileName, '.');

  {
    int n = 0;
    while (*tmp != '\0')
    {
      pripona[n] = tolower(*tmp);
      tmp++;
      n++;
    }
    pripona[n] = '\0';
    tmp = pripona;
  }

  /* Decoding */
  if(strcmp(tmp,".tga") == 0)
  {
    if ((r = decodeTga(fpIn, &buffer, &xRes, &yRes)) <= 0)
    {
      fclose(fpIn);
      return r;
    }
  }
  else
    if(strcmp(tmp,".pcx") == 0)
    {
      if ((r = decodePcx(fpIn, &buffer, &xRes, &yRes)) <= 0)
      {
        fclose(fpIn);
        return r;
      }
    }
    else
    {
      fclose(fpIn);
      /*fputs("Bad input file!\n", stderr);*/
      return makeTexUnknownFileExt;
    }

  fclose(fpIn);

  /* Creating texture */
  glGenTextures(1, TexID);
  glBindTexture(GL_TEXTURE_2D, *TexID);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
/*
  glTexImage2D(GL_TEXTURE_2D,
               0,
               GL_RGB,
               xRes,
               yRes,
               0,
               GL_RGB,
               GL_UNSIGNED_BYTE,
               buffer
               );
*/
  if ((r = gluBuild2DMipmaps(GL_TEXTURE_2D, 3, xRes, yRes, GL_RGB,
		    GL_UNSIGNED_BYTE, buffer)) != 0)
  {
    free(buffer);
    if (r==GLU_OUT_OF_MEMORY) return makeTexGLUOutOfMem;
    else return makeTexGLUError;
  }

  free(buffer);

  return makeTexSuccess;
}
#endif
