/*
 * The POG project
 * Copyright 2000 The POG project team
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 */



#include <string.h>
#include <assert.h>
#include "mygl.h"
#include "config.h"
#ifdef WINDOZE
#include <windows.h>
#endif
#pragma hdrstop // for C++ Builder compiler speed up

#include "gui.h"
#include "guiglut.h"
#include "Cockpit.h"
#include "Timers.h"
#include "Scene.h"
#include "SceneClasses.h"
#include "3DMath.h"
#include "Input.h"
#include "Log.h"
#include "guisdl.h"
#include "Cameras.h"
#include "InputDevices.h"

#include "setup.h"


#ifdef USE_GLUT
#define guiXxxInit guiGlutInit
#define guiXxxDone guiGlutDone
#define guiXxxMainLoop guiGlutMainLoop
#endif
#ifdef USE_SDL
#define guiXxxInit guiSdlInit
#define guiXxxDone guiSdlDone
#define guiXxxMainLoop guiSdlMainLoop
#endif


GuiInfo guiInfo;

static bool guiInitialized = false;


static void CheckHWAcceleration();
static int CheckGLExtensions();



int guiInit(int argc, char *argv[])
{
	if (!guiInitialized)
	{
		#define InitialWidth 640
		#define InitialHeight 480
		guiInfo.argc = argc;
		guiInfo.argv = argv;
		guiInfo.winWidth = InitialWidth;
		guiInfo.winHeight = InitialHeight;
		guiInfo.winWidth2 = InitialWidth>>1;
		guiInfo.winHeight2 = InitialHeight>>1;

		if (!guiXxxInit())  return 0;

		if (!CheckGLExtensions())  return 0;

		CheckHWAcceleration();

		glEnable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glDepthFunc(GL_LESS);
		glDisable(GL_DITHER);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_TEXTURE_2D);

		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glClearColor(0.0, 0.0, 0.2, 1.0);

		glDisable(GL_CULL_FACE);
//		glCullFace(GL_FRONT);

		guiInitialized = true;
		logInfo(modGUI, "GUI started successfully.");
	}

	return 1;
}


void guiDone()
{
	if (guiInitialized)
	{
		guiInitialized = false;
		guiXxxDone();
	}
}


void guiMainLoop(void)
{
	InputAcquire();
	SceneFreeze(false);
	guiXxxMainLoop();
}



void guiRenderScene(void)
{
#ifdef LOG_INFO_ENABLED
	static unsigned int FramesCounter = 0;
	if (FramesCounter >= 20)
	{
		logInfo(modGUI, "From now GUI and Cockpit modules will log only warninigs and errors to protect size of log file.");
		logLevelTable[modGUI] = logLevelWarnings;
		logLevelTable[modCockpit] = logLevelWarnings;
	}
	else
		++FramesCounter;
#endif

	InputUpdate();

	guiCheckForGLErrorsQ();

	if (ThisPlayer != NULL)
	{
		MatrixRotate(&ThisPlayer->Matrix, -InputAxes.RotLR, InputAxes.RotDU, InputAxes.Roll);
		MatrixMove(&ThisPlayer->Matrix, InputAxes.MovForward, InputAxes.MovLR, InputAxes.MovDU);
	}

	logInfo(modGUI, "Starting frame rendering.");

	MainCamera.Render();
	CockpitDraw();

	glFlush();

	if (guiCheckForGLErrors(modGUI, "Frame rendering terminated with error(s): %s"))
		logInfo(modGUI, "Frame rendering terminated successfully.");
}


void guiAfterRendering(void)
{
	frNextFrame();
	if (InputKeys.Pause)
		if (SceneFreezeStatus() == fsPaused) SceneFreeze(false);
		else SceneFreeze(true);
	SceneTimeTick(frLastFrameTime);
}


void guiWindowReshape(int width, int height)
{
	double ratio;

	logInfo(modGUI, "Window is reshaped to %ix%i.", width, height);

	guiInfo.winWidth = width;
	guiInfo.winHeight = height;
	guiInfo.winWidth2 = width >> 1;
	guiInfo.winHeight2 = height >> 1;

	ratio = (double)width / height;
	glViewport(0,0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glOrtho(0, width, 0, height, 1, 10);
	glFrustum(-1, 1, -1/ratio, 1/ratio, 1, 2300);
  //gluPerspective(60, ratio, 1, 1000);

	MouseWindowCallback();
}


void guiWindowMove(void)
{
	logInfo(modGUI, "Window is moved.");
	MouseWindowCallback();
}



static void CheckHWAcceleration(void)
{
#ifndef UNIX
	PIXELFORMATDESCRIPTOR pfd;
	HDC hdc = wglGetCurrentDC();
	int ipf = GetPixelFormat(hdc);
	if (ipf>=1)
	{
	#ifdef USE_GLUT
		if (glutGameModeGet(GLUT_GAME_MODE_ACTIVE))
		{
			int width = glutGameModeGet(GLUT_GAME_MODE_WIDTH);
			int height = glutGameModeGet(GLUT_GAME_MODE_HEIGHT);
			int bpp = glutGameModeGet(GLUT_GAME_MODE_PIXEL_DEPTH);
			int refreshrate = glutGameModeGet(GLUT_GAME_MODE_REFRESH_RATE);

			if (DescribePixelFormat(hdc, ipf, sizeof(pfd), &pfd))
			{
				if ((pfd.dwFlags & PFD_GENERIC_FORMAT) && !(pfd.dwFlags & PFD_GENERIC_ACCELERATED))
					logInfo(modGUI, "GameMode started successfuly (%dx%d:%d@%d) with software rendering.", width, height, bpp, refreshrate);
				else
					logInfo(modGUI, "GameMode started successfuly (%dx%d:%d@%d) with hardware acceleration.", width, height, bpp, refreshrate);
			}
			else
				logError(modGUI, "GameMode started successfuly (%dx%d:%d@%d), but can't get rendering type.", width, height, bpp, refreshrate);
		}
		else
		{
			int width = glutGet(GLUT_WINDOW_WIDTH);
			int height = glutGet(GLUT_WINDOW_HEIGHT);
			int bpp = glutGet(GLUT_WINDOW_BUFFER_SIZE);
			int depth = glutGet(GLUT_WINDOW_DEPTH_SIZE);

			if (DescribePixelFormat(hdc, ipf, sizeof(pfd), &pfd))
			{
				if ((pfd.dwFlags & PFD_GENERIC_FORMAT) && !(pfd.dwFlags & PFD_GENERIC_ACCELERATED))
					logInfo(modGUI, "Window mode started successfuly (%dx%d:%dz%d) with software rendering.", width, height, bpp, depth);
				else
					logInfo(modGUI, "Window mode started successfuly (%dx%d:%dz%d) with hardware acceleration.", width, height, bpp, depth);
			}
			else
				logError(modGUI, "Window mode started successfuly (%dx%d:%dz%d), but can't get rendering type.", width, height, bpp, depth);
		}
	#else
		if (DescribePixelFormat(hdc, ipf, sizeof(pfd), &pfd))
		{
			if ((pfd.dwFlags & PFD_GENERIC_FORMAT) && !(pfd.dwFlags & PFD_GENERIC_ACCELERATED))
				logInfo(modGUI, "SDL started successfuly with software rendering.");
			else
				logInfo(modGUI, "SDL started successfuly with hardware acceleration.");
		}
		else
			logError(modGUI, "SDL started successfuly, but can't get rendering type.");
	#endif
	}
#endif
}


static int CheckGLExtensions()
{
	logInfo(modGUI, "OpenGL Driver: Vendor=\"%s\" Renderer=\"%s\" Version=\"%s.\"",
		glGetString(GL_VENDOR), glGetString(GL_RENDERER), glGetString(GL_VERSION));
	logInfo(modGUI, "OpenGL Driver extensions: %s.", glGetString(GL_EXTENSIONS));

	{
		const char *A;
		// major version
		const char *S = (const char*)glGetString(GL_VERSION);
		const char *T = strstr(S, ".");
		int V = 0;
		int exp = 1;
		for (A=S; A<T; ++A)
		{
			if ((*A >= '0') && (*A <= '9'))
			{
				V = (*A - '0') * exp;
				exp *= 10;
			}
			else
			{
				logError(modGUI, "Bad OpenGL version: \"%s\".", glGetString(GL_VERSION));
				return 0;
			}
		}
		if (V<1)
		{
			logError(modGUI, "Bad OpenGL version: \"%s\", need version 1.1 or greater.", glGetString(GL_VERSION));
			return 0;
		}
		else
			if (V == 1)
			{
				// minor version
				S = T+1;
				A = strstr(S, " ");
				T = strstr(S, ".");
				if (A == NULL)
					if (T == NULL)
					{
						T = S;
						while (*T!=0) ++T;
					}
					else;
				else
					if ((T > A) || (T == NULL))
						T = A;

				V = 0;
				exp = 1;
				for (A=S; A<T; ++A)
				{
					if ((*A >= '0') && (*A <= '9'))
					{
						V = (*A - '0') * exp;
						exp *= 10;
					}
					else
					{
						logError(modGUI, "Bad OpenGL version: \"%s\".", glGetString(GL_VERSION));
						return 0;
					}
				}
			if (V<1)
			{
				logError(modGUI, "Bad OpenGL version: \"%s\", need version 1.1 or greater.", glGetString(GL_VERSION));
				return 0;
			}
		}
	}
	return 1;
}


static const char errInvalidEnum[]   = "GL_INVALID_ENUM";
static const char errInvalidValue[]  = "GL_INVALID_VALUE";
static const char errInvalidOperation[] = "GL_INVALID_OPERATION";
static const char errStackOverflow[] = "GL_STACK_OVERFLOW";
static const char errStackUnderflow[]= "GL_STACK_UNDERFLOW";
static const char errOutOfMemory[]   = "GL_OUT_OF_MEMORY";
static const char errNoError[] = "GL_NO_ERROR";
static const char errUnknownError[]  = "unknown error";

static const char *ErrorToString(GLenum E)
{
	switch (E)
	{
		case GL_INVALID_ENUM   : return &errInvalidEnum[0];
		case GL_INVALID_VALUE  : return &errInvalidValue[0];
		case GL_INVALID_OPERATION: return &errInvalidOperation[0];
		case GL_STACK_OVERFLOW : return &errStackOverflow[0];
		case GL_STACK_UNDERFLOW: return &errStackUnderflow[0];
		case GL_OUT_OF_MEMORY  : return &errOutOfMemory[0];
		case GL_NO_ERROR : return &errNoError[0];
		default : return &errUnknownError[0];
	}
}


int guiCheckForGLErrors(logModule M,char *ErrorString)
{
	GLenum R,LR;

	R = glGetError();
	if (R == GL_NO_ERROR)
	{
		return 1;
	}
	else
	{
		logError(M, ErrorString, ErrorToString(R));
		LR = R;
		R = glGetError();
		while (R!=GL_NO_ERROR)
		{
			if (R==LR)
			{
				logFatalError(M, "OpenGL turned to unstable state. Can't reset %i error flag.", ErrorToString(R));
                                //				guiForcedTerminate(); !!!!!!!!!!!!!!!!!!!!
			}
			logError(M, "Next error: %i", ErrorToString(R));
			LR = R;
			R = glGetError();
		}
		return 0;
	}
}

